# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration to train a fairness aware classifier on the adult dataset."""

import ml_collections


def get_config():
  """Return a ConfigDict."""
  config = ml_collections.ConfigDict()
  config.folder = '/tmp/adult_dataset/'
  config.training_filename = 'adult.data'
  config.test_filename = 'adult.test'
  config.info_filename = 'adult.names'
  config.protected = 'sex'

  config.batch_size = 256
  config.num_epochs = 20
  config.embed_dim = 16
  config.hidden_layers = (64, 64)
  config.learning_rate = 1e-4

  config.epsilon = 1e-3
  config.quantization = 16
  config.num_groups = 2
  config.fair_weight = 1.0
  return config



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Implements a geometry class for points supported on a cartesian product."""
import itertools
from typing import Any, List, NoReturn, Optional, Sequence, Tuple

import jax
import jax.numpy as jnp
import numpy as np

from ott.geometry import costs, geometry, pointcloud
from ott.math import utils

__all__ = ["Grid"]


@jax.tree_util.register_pytree_node_class
class Grid(geometry.Geometry):
  r"""Class describing the geometry of points taken in a Cartesian product.

  This class implements a geometry in which probability measures are supported
  on a :math:`d`-dimensional cartesian grid, a cartesian product of :math:`d`
  lists of values, each list being itself of size :math:`n_i`.

  The transportation cost between points in the grid is assumed to be separable,
  namely a sum of coordinate-wise cost functions, as in:

  .. math::

    cost(x,y) = \sum_{i=1}^d cost_i(x_i, y_i)

  where :math:`cost_i`: R x R → R.

  In such a regime, and despite the fact that the total number :math:`n_{total}`
  of points in the grid is exponential :math:`d` (namely :math:`\prod_i n_i`),
  applying a kernel in the context of regularized optimal transport can be
  carried out in time that is of the order of :math:`n_{total}^{(1+1/d)}` using
  convolutions, either in the original domain or log-space domain. This class
  precomputes :math:`d` :math:`n_i` x :math:`n_i` cost matrices (one per
  dimension) and implements these two operations by carrying out these
  convolutions one dimension at a time.

  Args:
    x : list of arrays of varying sizes, describing the locations of the grid.
      Locations are provided as a list of arrays, that is :math:`d`
      vectors of (possibly varying) size :math:`n_i`. The resulting grid
      is the Cartesian product of these vectors.
    grid_size: tuple of integers describing grid sizes, namely
      :math:`(n_1,...,n_d)`. This will only be used if x is None.
      In that case the grid will be assumed to lie in the hypercube
      :math:`[0,1]^d`, with the :math:`d` dimensions, described as points
      regularly sampled in [0,1].
    cost_fns: a sequence of :math:`d` costs.CostFn's, each being a cost taking
      two reals as inputs to output a real number.
    num_a: total size of grid. This parameters will be computed from other
      inputs and used in the flatten/unflatten functions.
    grid_dimension: dimension of grid. This parameters will be computed from
      other inputs and used in the flatten/unflatten functions.
    kwargs: other optional parameters to be passed on to superclass
      initializer, notably those related to epsilon regularization.
  """

  def __init__(
      self,
      x: Optional[Sequence[jnp.ndarray]] = None,
      grid_size: Optional[Sequence[int]] = None,
      cost_fns: Optional[Sequence[costs.CostFn]] = None,
      num_a: Optional[int] = None,
      grid_dimension: Optional[int] = None,
      **kwargs: Any,
  ):
    if (
        grid_size is not None and x is not None and num_a is not None and
        grid_dimension is not None
    ):
      self.grid_size = tuple(map(int, grid_size))
      self.x = x
      self.num_a = num_a
      self.grid_dimension = grid_dimension
    elif x is not None:
      self.x = x
      self.grid_size = tuple(xs.shape[0] for xs in x)
      self.num_a = np.prod(np.array(self.grid_size))
      self.grid_dimension = len(self.x)
    elif grid_size is not None:
      self.grid_size = tuple(map(int, grid_size))
      self.x = tuple(jnp.linspace(0, 1, n) for n in self.grid_size)
      self.num_a = np.prod(np.array(grid_size))
      self.grid_dimension = len(self.grid_size)
    else:
      raise ValueError('Input either grid_size t-uple or grid locations x.')

    if cost_fns is None:
      cost_fns = [costs.SqEuclidean()]
    self.cost_fns = cost_fns
    self.kwargs = {
        'num_a': self.num_a,
        'grid_size': self.grid_size,
        'grid_dimension': self.grid_dimension
    }

    super().__init__(**kwargs)

  @property
  def cost_matrices(self) -> List[jnp.ndarray]:
    """Cost matrices along each dimension of the grid."""
    cost_matrices = []
    for dimension, cost_fn in itertools.zip_longest(
        range(self.grid_dimension), self.cost_fns, fillvalue=self.cost_fns[-1]
    ):
      x_values = self.x[dimension][:, jnp.newaxis]
      cost_matrices.append(
          pointcloud.PointCloud(x_values, cost_fn=cost_fn).cost_matrix
      )
    return cost_matrices

  @property
  def kernel_matrices(self) -> List[jnp.ndarray]:
    """Kernel matrices along each dimension of the grid."""
    kernel_matrices = []
    for cost_matrix in self.cost_matrices:
      kernel_matrices.append(jnp.exp(-cost_matrix / self.epsilon))
    return kernel_matrices

  @property
  def median_cost_matrix(self) -> NoReturn:
    """Not implemented."""
    raise NotImplementedError('Median cost not implemented for grids.')

  @property
  def shape(self) -> Tuple[int, int]:
    return self.num_a, self.num_a

  @property
  def is_symmetric(self) -> bool:
    return True

  # Reimplemented functions to be used in regularized OT
  def apply_lse_kernel(
      self,
      f: jnp.ndarray,
      g: jnp.ndarray,
      eps: float,
      vec: Optional[jnp.ndarray] = None,
      axis: int = 0
  ) -> jnp.ndarray:
    """Apply grid kernel in log space. See notes in parent class for use case.

    Reshapes vector inputs below as grids, applies kernels onto each slice, and
    then expands the outputs as vectors.

    More implementation details in :cite:`schmitz:18`.

    Args:
      f: jnp.ndarray, a vector of potentials
      g: jnp.ndarray, a vector of potentials
      eps: float, regularization strength
      vec: jnp.ndarray, if needed, a vector onto which apply the kernel weighted
        by f and g.
      axis: axis (0 or 1) along which summation should be carried out.

    Returns:
      a vector, the result of kernel applied in lse space onto vec.
    """
    f, g = jnp.reshape(f, self.grid_size), jnp.reshape(g, self.grid_size)

    if vec is not None:
      vec = jnp.reshape(vec, self.grid_size)

    if axis == 0:
      f, g = g, f

    for dimension in range(self.grid_dimension):
      g, vec = self._apply_lse_kernel_one_dimension(dimension, f, g, eps, vec)
      g -= jnp.where(jnp.isfinite(f), f, 0)
    if vec is None:
      vec = jnp.array(1.0)
    return g.ravel(), vec.ravel()

  def _apply_lse_kernel_one_dimension(self, dimension, f, g, eps, vec=None):
    """Permute axis & apply the kernel on a single slice."""
    indices = np.arange(self.grid_dimension)
    indices[dimension], indices[0] = 0, dimension
    f, g = jnp.transpose(f, indices), jnp.transpose(g, indices)
    centered_cost = (
        f[:, jnp.newaxis, ...] + g[jnp.newaxis, ...] - jnp.expand_dims(
            self.cost_matrices[dimension],
            axis=tuple(range(2, 1 + self.grid_dimension))
        )
    ) / eps

    if vec is not None:
      vec = jnp.transpose(vec, indices)
      softmax_res, softmax_sgn = utils.logsumexp(
          centered_cost, b=vec, axis=1, return_sign=True
      )
      return eps * jnp.transpose(softmax_res,
                                 indices), jnp.transpose(softmax_sgn, indices)
    softmax_res = utils.logsumexp(centered_cost, axis=1)
    return eps * jnp.transpose(softmax_res, indices), None

  def _apply_cost_to_vec(
      self, vec: jnp.ndarray, axis: int = 0, fn=None
  ) -> jnp.ndarray:
    r"""Apply grid's cost matrix (without instantiating it) to a vector.

    The `apply_cost` operation on grids rests on the following identity.
    If it were to be cast as a [num_a, num_a] matrix, the corresponding cost
    matrix :math:`C` would be a sum of grid_dimension matrices, each of the form
    (here for the j-th slice)
    :math:`\tilde{C}_j : = 1_{n_1} \otimes \dots \otimes C_j \otimes 1_{n_d}`
    where each :math:`1_{n}` is the :math:`n\times n` square matrix full of 1's.

    Applying :math:`\tilde{C}_j` on a vector grid consists in carrying a tensor
    multiplication on the dimension of that vector reshaped as a grid, followed
    by a summation on all other axis while keeping dimensions. That identity is
    a generalization of the formula
    :math:`(1_{n} \otimes A) vec(X) = vec( A X 1_n)`
    where the last multiplication by the matrix of ones is equivalent to
    summation while keeping dimensions.

    Args:
      vec: jnp.ndarray, flat vector of total size prod(grid_size).
      axis: axis 0 if applying transpose costs, 1 if using the original cost.
      fn: function optionally applied to cost matrix element-wise, before the
        dot product.

    Returns:
      A jnp.ndarray corresponding to cost x matrix
    """
    vec = jnp.reshape(vec, self.grid_size)
    accum_vec = jnp.zeros_like(vec)
    indices = list(range(1, self.grid_dimension))
    for dimension, cost in enumerate(self.cost_matrices):
      ind = indices.copy()
      ind.insert(dimension, 0)
      if axis == 0:
        cost = cost.T
      accum_vec += jnp.sum(
          jnp.tensordot(cost, vec, axes=([0], [dimension])),
          axis=indices,
          keepdims=True
      ).transpose(ind)
    return accum_vec.ravel()

  def apply_kernel(
      self,
      scaling: jnp.ndarray,
      eps: Optional[float] = None,
      axis: Optional[int] = None
  ) -> jnp.ndarray:
    """Apply grid kernel on scaling vector.

    See notes in parent class for use.

    Reshapes scaling vector as a grid, applies kernels onto each slice, and
    then ravels backs the output as a vector.

    More implementation details in :cite:`schmitz:18`,

    Args:
      scaling: jnp.ndarray, a vector of scaling (>0) values.
      eps: float, regularization strength
      axis: axis (0 or 1) along which summation should be carried out.

    Returns:
      a vector, the result of kernel applied onto scaling.
    """
    scaling = jnp.reshape(scaling, self.grid_size)
    indices = list(range(1, self.grid_dimension))
    for dimension, kernel in enumerate(self.kernel_matrices):
      ind = indices.copy()
      ind.insert(dimension, 0)
      kernel = kernel if eps is None else kernel ** (self.epsilon / eps)
      scaling = jnp.tensordot(
          kernel, scaling, axes=([0], [dimension])
      ).transpose(ind)
    return scaling.ravel()

  def transport_from_potentials(
      self, f: jnp.ndarray, g: jnp.ndarray, axis: int = 0
  ) -> NoReturn:
    """Not implemented, use :meth:`apply_transport_from_potentials` instead."""
    raise ValueError(
        'Grid geometry cannot instantiate a transport matrix, use',
        ' apply_transport_from_potentials(...) if you wish to ',
        ' apply the transport matrix to a vector, or use a point '
        ' cloud geometry instead'
    )

  def transport_from_scalings(
      self, f: jnp.ndarray, g: jnp.ndarray, axis: int = 0
  ) -> NoReturn:
    """Not implemented, use :meth:`apply_transport_from_scalings` instead."""
    raise ValueError(
        'Grid geometry cannot instantiate a transport matrix, use ',
        'apply_transport_from_scalings(...) if you wish to ',
        'apply the transport matrix to a vector, or use a point '
        'cloud geometry instead.'
    )

  def subset(
      self, src_ixs: Optional[jnp.ndarray], tgt_ixs: Optional[jnp.ndarray]
  ) -> NoReturn:
    """Not implemented."""
    raise NotImplementedError("Subsetting is not implemented for grids.")

  def mask(
      self,
      src_mask: Optional[jnp.ndarray],
      tgt_mask: Optional[jnp.ndarray],
      mask_value: float = 0.,
  ) -> NoReturn:
    """Not implemented."""
    raise NotImplementedError("Masking is not implemented for grids.")

  @classmethod
  def prepare_divergences(
      cls,
      *args: Any,
      static_b: bool = False,
      **kwargs: Any
  ) -> Tuple["Grid", ...]:
    """Instantiate the geometries used for a divergence computation."""
    grid_size = kwargs.pop('grid_size', None)
    x = kwargs.pop('x', args)

    sep_grid = cls(x=x, grid_size=grid_size, **kwargs)
    size = 2 if static_b else 3
    return tuple(sep_grid for _ in range(size))

  @property
  def dtype(self) -> jnp.dtype:
    return self.x[0].dtype

  def tree_flatten(self):
    return (self.x, self.cost_fns, self._epsilon), self.kwargs

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    return cls(
        x=children[0], cost_fns=children[1], epsilon=children[2], **aux_data
    )



import functools
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp
from typing_extensions import Literal

from ott.geometry import costs, geometry, pointcloud, segment
from ott.math import utils as mu
from ott.problems.linear import barycenter_problem
from ott.problems.quadratic import quadratic_costs, quadratic_problem

__all__ = ["GWBarycenterProblem"]


# TODO(michalk8): better abstraction (common superclass for Wasserstein bary)
@jax.tree_util.register_pytree_node_class
class GWBarycenterProblem(barycenter_problem.BarycenterProblem):
  """(Fused) Gromov-Wasserstein barycenter problem :cite:`peyre:16,vayer:19`.

  Args:
    y: Array of shape ``[num_total_points, ndim]`` merging the points of all
      measures. Alternatively, already segmented array of shape
      ``[num_measures, max_measure_size, ndim]`` can be passed.
      See also :func:`~ott.geometry.segment.segment_point_cloud`.
    b: Array of shape ``[num_total_points,]`` containing the weights of all
      the points within the measures that define the barycenter problem.
      Same as ``y``, pre-segmented array of weights of shape
      ``[num_measures, max_measure_size]`` can be passed.
      If ``y`` is already pre-segmented, this array must be passed.
    weights: Array of shape ``[num_measures,]`` containing the weights of the
      barycenter problem.
    costs: Alternative to ``y``, an array of shape
      ``[num_measures, max_measure_size, max_measure_size]`` that defines padded
      cost matrices for each measure. Used in the quadratic term.
      Only one of ``y`` and ``cost`` can be specified.
    y_fused: Array of shape ``[num_total_points, ndim_fused]`` containing
      the data of the points of all measures used to define the linear term
      in the fused case. Same as ``y``, it can be specified as a pre-segmented
      array of shape ``[num_measures, max_measure_size, ndim_fused]``.
    gw_loss: Gromov-Wasserstein loss.
    fused_penalty: Multiplier of the linear term. Only used when
      ``y_fused != None``.
    scale_cost: Scaling of cost matrices passed to geometries.
    kwargs: Keyword arguments for
      :class:`~ott.problems.linear.barycenter_problem.BarycenterProblem`.
  """

  def __init__(
      self,
      y: Optional[jnp.ndarray] = None,
      b: Optional[jnp.ndarray] = None,
      weights: Optional[jnp.ndarray] = None,
      costs: Optional[jnp.ndarray] = None,
      y_fused: Optional[jnp.ndarray] = None,
      fused_penalty: float = 1.0,
      gw_loss: Literal['sqeucl', 'kl'] = 'sqeucl',
      scale_cost: Union[int, float, Literal["mean", "max_cost"]] = 1.0,
      **kwargs: Any,
  ):
    assert y is None or costs is None, "Cannot specify both `y` and `costs`."
    y = y if costs is None else costs

    super().__init__(y=y, b=b, weights=weights, **kwargs)

    self._y_fused = y_fused
    self.fused_penalty = fused_penalty
    self._loss_name = gw_loss
    self.scale_cost = scale_cost
    self._y_as_costs = costs is not None

    if self._y_as_costs:
      # (num_measures, max_measure_size, max_measure_size)
      _, n, m = self._y.shape
      assert n == m, "Cost matrices must be square."
    if self.is_fused:
      seg_y = self._is_segmented
      seg_fused = self._y_fused.ndim == 3
      if seg_y and seg_fused:
        # (num_measures, max_measure_size, ndim_fused)
        # (num_measures, max_measure_size, ndim)
        assert self._y_fused.shape[:2] == self._y.shape[:2]
      if not seg_y and not seg_fused:
        # (num_total_points, ndim_fused), (num_total_points, ndim)
        assert self._y_fused.shape[0] == self._y.shape[0]
      # TODO(michalk8): in the future, consider checking the other 2 cases
      # using `segmented_y` and `segmented_y_fused`?

  def update_barycenter(
      self, transports: jnp.ndarray, a: jnp.ndarray
  ) -> jnp.ndarray:
    """Update the barycenter cost matrix.

    Uses the eq. 14 and 15 of :cite:`peyre:16`.

    Args:
      transports: Transport maps of shape
        ``[num_measures, bar_size, max_measure_size]``.
      a: Barycenter weights of shape ``[bar_size,]``.

    Returns:
      Update cost matrix of shape ``[bar_size, bar_size]``.
    """

    @functools.partial(jax.vmap, in_axes=[0, 0, 0, None])
    def project(
        y: jnp.ndarray,
        b: jnp.ndarray,
        transport: jnp.ndarray,
        fn: Optional[quadratic_costs.Loss],
    ) -> jnp.ndarray:
      geom = self._create_y_geometry(y, mask=b > 0.)
      fn, lin = (None, True) if fn is None else (fn.func, fn.is_linear)

      tmp = geom.apply_cost(
          transport.T,
          axis=0,
          fn=fn,
          is_linear=lin,
      )
      return transport @ tmp

    fn = None if self._loss_name == 'sqeucl' else self.gw_loss.h2
    y, b = self.segmented_y_b
    weights = self.weights[:, None, None]

    barycenter = jnp.sum(weights * project(y, b, transports, fn), axis=0)
    inv_a = jnp.where(a > 0, 1.0 / a, 1.0)
    barycenter = (barycenter * inv_a[None, :]) * inv_a[:, None]

    # TODO(michalk8): in future, use `isinstanceof(self.gw_loss, ...)`
    # once refactoring has been done
    if self._loss_name == 'kl':
      barycenter = jnp.exp(barycenter)
    return barycenter

  def update_features(self, transports: jnp.ndarray,
                      a: jnp.ndarray) -> Optional[jnp.ndarray]:
    """Update the barycenter features in the fused case :cite:`vayer:19`.

    Uses :cite:`cuturi:14` eq. 8, and is implemented only
    for the :class:`~ott.geometry.costs.SqEuclidean` cost.

    Args:
      transports: Transport maps of shape
        ``[num_measures, bar_size, max_measure_size]``.
      a: Barycenter weights of shape ``[bar_size,]``.

    Returns:
      Updated features of shape ``[bar_size, ndim_fused]``.
    """
    if not self.is_fused:
      raise RuntimeError(
          "Updating features is available only in the fused case."
      )

    y_fused = self.segmented_y_fused
    weights = self.weights[:, None, None]
    inv_a = jnp.where(a > 0, 1.0 / a, 1.0)
    transports = transports * inv_a[None, :, None]

    if self._loss_name == "sqeucl":
      cost_fn = costs.SqEuclidean()
      return jnp.sum(
          weights * mu.barycentric_projection(transports, y_fused, cost_fn),
          axis=0
      )
    raise NotImplementedError(self._loss_name)

  def _create_bary_geometry(
      self,
      cost_matrix: jnp.ndarray,
      mask: Optional[jnp.ndarray] = None
  ) -> geometry.Geometry:
    return geometry.Geometry(
        cost_matrix=cost_matrix,
        src_mask=mask,
        tgt_mask=mask,
        epsilon=self.epsilon,
        scale_cost=self.scale_cost
    )

  def _create_y_geometry(
      self,
      y: jnp.ndarray,
      mask: Optional[jnp.ndarray] = None
  ) -> geometry.Geometry:
    if self._y_as_costs:
      assert y.shape[0] == y.shape[1], y.shape
      return geometry.Geometry(
          y,
          epsilon=self.epsilon,
          scale_cost=self.scale_cost,
          src_mask=mask,
          tgt_mask=mask
      )
    return pointcloud.PointCloud(
        y,
        epsilon=self.epsilon,
        scale_cost=self.scale_cost,
        cost_fn=self.cost_fn,
        src_mask=mask,
        tgt_mask=mask
    )

  def _create_fused_geometry(
      self,
      x: jnp.ndarray,
      y: jnp.ndarray,
      src_mask: Optional[jnp.ndarray] = None,
      tgt_mask: Optional[jnp.ndarray] = None
  ) -> pointcloud.PointCloud:
    return pointcloud.PointCloud(
        x,
        y,
        cost_fn=self.cost_fn,
        epsilon=self.epsilon,
        scale_cost=self.scale_cost,
        src_mask=src_mask,
        tgt_mask=tgt_mask
    )

  def _create_problem(
      self,
      state: 'GWBarycenterState',  # noqa: F821
      y: jnp.ndarray,
      b: jnp.ndarray,
      f: Optional[jnp.ndarray] = None
  ) -> quadratic_problem.QuadraticProblem:
    # TODO(michalk8): in future, mask in the problem for convenience?
    bary_mask = state.a > 0.
    y_mask = b > 0.

    geom_xx = self._create_bary_geometry(state.cost, mask=bary_mask)
    geom_yy = self._create_y_geometry(y, mask=y_mask)
    if self.is_fused:
      assert f is not None
      assert state.x.shape[1] == f.shape[1]
      geom_xy = self._create_fused_geometry(
          state.x, f, src_mask=bary_mask, tgt_mask=y_mask
      )
    else:
      geom_xy = None

    return quadratic_problem.QuadraticProblem(
        geom_xx=geom_xx,
        geom_yy=geom_yy,
        geom_xy=geom_xy,
        a=state.a,
        b=b,
        fused_penalty=self.fused_penalty,
    )

  @property
  def is_fused(self) -> bool:
    """Whether the problem is fused."""
    return self._y_fused is not None

  @property
  def segmented_y_fused(self) -> Optional[jnp.ndarray]:
    """Feature array of shape ``[num_measures, max_measure_size, ndim_fused]`` \
    used in the fused case."""
    if not self.is_fused or self._y_fused.ndim == 3:
      return self._y_fused
    y_fused, _ = segment.segment_point_cloud(
        x=self._y_fused,
        padding_vector=self.cost_fn._padder(self.ndim_fused),
        **self._kwargs
    )
    return y_fused

  @property
  def ndim(self) -> Optional[int]:
    return None if self._y_as_costs else self._y.shape[-1]

  @property
  def ndim_fused(self) -> Optional[int]:
    """Number of dimensions of the fused term."""
    return self._y_fused.shape[-1] if self.is_fused else None

  @property
  def gw_loss(self) -> quadratic_costs.GWLoss:
    """Gromov-Wasserstein loss."""
    # TODO(michalk8): custom losses would require inverting some fns;
    # `https://jax.readthedocs.io/en/latest/notebooks/ some fns;
    # Writing_custom_interpreters_in_Jax.html#your-first-interpreter-invert`
    # might be useful
    if self._loss_name == 'sqeucl':
      return quadratic_costs.make_square_loss()
    if self._loss_name == 'kl':
      return quadratic_costs.make_kl_loss()
    raise NotImplementedError(
        f"Loss `{self._loss_name}` is not yet implemented."
    )

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:
    (y, b, weights), aux = super().tree_flatten()
    if self._y_as_costs:
      children = [None, b, weights, y]
    else:
      children = [y, b, weights, None]
    aux['fused_penalty'] = self.fused_penalty
    aux['gw_loss'] = self._loss_name
    aux['scale_cost'] = self.scale_cost
    return children + [self._y_fused], aux

  @classmethod
  def tree_unflatten(
      cls, aux_data: Dict[str, Any], children: Sequence[Any]
  ) -> "GWBarycenterProblem":
    y, b, weights, costs, y_fused = children
    return cls(
        y=y, b=b, weights=weights, costs=costs, y_fused=y_fused, **aux_data
    )



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Plotting utils."""

from typing import List, Optional, Sequence, Union

import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import scipy
from matplotlib import animation

from ott.geometry import pointcloud
from ott.tools import transport


def bidimensional(x: jnp.ndarray, y: jnp.ndarray):
  """Apply PCA to reduce to bimensional data."""
  if x.shape[1] < 3:
    return x, y

  u, s, _ = scipy.sparse.linalg.svds(jnp.concatenate([x, y], axis=0), k=2)
  proj = u * s
  k = x.shape[0]
  return proj[:k], proj[k:]


class Plot:
  """Plot an optimal transport map between two point clouds.

  It enables to either plot or update a plot in a single object, offering the
  possibilities to create animations as matplotlib.animation.FuncAnimation,
  which can in turned be saved to disk at will. There are two design principles
  here: 1) we do not rely on saving to/loading from disk to create animations
  2) we try as much as possible to disentangle the transport problem(s) from
  its visualization(s).
  """

  def __init__(
      self,
      fig: Optional[plt.Figure] = None,
      ax: Optional[plt.Axes] = None,
      cost_threshold: float = -1.0,  # should be negative for animations.
      scale: int = 200,
      show_lines: bool = True,
      cmap: str = 'cool'
  ):
    if ax is None and fig is None:
      fig, ax = plt.subplots()
    elif fig is None:
      fig = plt.gcf()
    elif ax is None:
      ax = plt.gca()
    self.fig = fig
    self.ax = ax
    self._show_lines = show_lines
    self._lines = []
    self._points_x = None
    self._points_y = None
    self._threshold = cost_threshold
    self._scale = scale
    self._cmap = cmap

  def _scatter(self, ot: transport.Transport):
    """Compute the position and scales of the points on a 2D plot."""
    if not isinstance(ot.geom, pointcloud.PointCloud):
      raise ValueError('So far we only plot PointCloud geometry.')

    x, y = ot.geom.x, ot.geom.y
    a, b = ot.a, ot.b
    x, y = bidimensional(x, y)
    scales_x = a * self._scale * a.shape[0]
    scales_y = b * self._scale * b.shape[0]
    return x, y, scales_x, scales_y

  def _mapping(self, x: jnp.ndarray, y: jnp.ndarray, matrix: jnp.ndarray):
    """Compute the lines representing the mapping between the 2 point clouds."""
    u, v = jnp.where(matrix > self._threshold)
    c = matrix[jnp.where(matrix > self._threshold)]
    xy = jnp.concatenate([x[u], y[v]], axis=-1)
    result = []
    for i in range(xy.shape[0]):
      strength = jnp.max(jnp.array(matrix.shape)) * c[i]
      result.append((xy[i, [0, 2]], xy[i, [1, 3]], strength))
    return result

  def __call__(self, ot: transport.Transport) -> List[plt.Artist]:
    """Plot 2-D couplings. Projects via PCA if data is higher dimensional."""
    x, y, sx, sy = self._scatter(ot)
    self._points_x = self.ax.scatter(
        *x.T, s=sx, edgecolors='k', marker='o', label='x'
    )
    self._points_y = self.ax.scatter(
        *y.T, s=sy, edgecolors='k', marker='X', label='y'
    )
    self.ax.legend(fontsize=15)
    if not self._show_lines:
      return []

    lines = self._mapping(x, y, ot.matrix)
    cmap = plt.get_cmap(self._cmap)
    self._lines = []
    for start, end, strength in lines:
      line, = self.ax.plot(
          start,
          end,
          linewidth=0.5 + 4 * strength,
          color=cmap(strength),
          zorder=0,
          alpha=0.7
      )
      self._lines.append(line)
    return [self._points_x, self._points_y] + self._lines

  def update(self, ot: transport.Transport) -> List[plt.Artist]:
    """Update a plot with a transport.Transport instance."""
    x, y, _, _ = self._scatter(ot)
    self._points_x.set_offsets(x)
    self._points_y.set_offsets(y)
    if not self._show_lines:
      return []

    new_lines = self._mapping(x, y, ot.matrix)
    cmap = plt.get_cmap(self._cmap)
    for line, new_line in zip(self._lines, new_lines):
      start, end, strength = new_line
      line.set_data(start, end)
      line.set_linewidth(0.5 + 4 * strength)
      line.set_color(cmap(strength))

    # Maybe add new lines to the plot.
    num_lines = len(self._lines)
    num_to_plot = len(new_lines) if self._show_lines else 0
    for i in range(num_lines, num_to_plot):
      start, end, strength = new_lines[i]
      line, = self.ax.plot(
          start,
          end,
          linewidth=0.5 + 4 * strength,
          color=cmap(strength),
          zorder=0,
          alpha=0.7
      )
      self._lines.append(line)

    self._lines = self._lines[:num_to_plot]  # Maybe remove some
    return [self._points_x, self._points_y] + self._lines

  def animate(
      self,
      transports: Sequence[transport.Transport],
      frame_rate: float = 10.0
  ) -> animation.FuncAnimation:
    """Make an animation from several transport.Transport."""
    self(transports[0])
    return animation.FuncAnimation(
        self.fig,
        lambda i: self.update(transports[i]),
        np.arange(0, len(transports)),
        init_func=lambda: self.update(transports[0]),
        interval=1000 / frame_rate,
        blit=True
    )


def _barycenters(
    ax: plt.Axes,
    y: jnp.ndarray,
    a: jnp.ndarray,
    b: jnp.ndarray,
    matrix: jnp.ndarray,
    scale: int = 200
):
  """Plot 2-D sinkhorn barycenters."""
  sa, sb = jnp.min(a) / scale, jnp.min(b) / scale
  ax.scatter(*y.T, s=b / sb, edgecolors='k', marker='X', label='y')
  tx = 1 / a[:, None] * jnp.matmul(matrix, y)
  ax.scatter(*tx.T, s=a / sa, edgecolors='k', marker='X', label='T(x)')
  ax.legend(fontsize=15)


def barycentric_projections(
    arg: Union[transport.Transport, jnp.ndarray],
    a: jnp.ndarray = None,
    b: jnp.ndarray = None,
    matrix: jnp.ndarray = None,
    ax: Optional[plt.Axes] = None,
    **kwargs
):
  """Plot the barycenters, from the Transport object or from arguments."""
  if ax is None:
    _, ax = plt.subplots(1, 1, figsize=(8, 5))

  if isinstance(arg, transport.Transport):
    ot = arg
    return _barycenters(ax, ot.geom.y, ot.a, ot.b, ot.matrix, **kwargs)

  if matrix is None:
    raise ValueError('The `matrix` argument cannot be None.')

  a = jnp.ones(matrix.shape[0]) / matrix.shape[0] if a is None else a
  b = jnp.ones(matrix.shape[1]) / matrix.shape[1] if b is None else b
  return _barycenters(ax, arg, a, b, matrix, **kwargs)



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Several cost/norm functions for relevant vector types."""
import abc
import functools
import math
from typing import Any, Callable, Optional, Tuple, Union

import jax
import jax.numpy as jnp

from ott.math import fixed_point_loop, matrix_square_root

__all__ = [
    "PNorm", "SqPNorm", "Euclidean", "SqEuclidean", "Cosine", "Bures",
    "UnbalancedBures"
]


@jax.tree_util.register_pytree_node_class
class CostFn(abc.ABC):
  """A generic cost function, taking two vectors as input.

  Cost functions evaluate a function on a pair of inputs. For convenience,
  that function is split into two norms -- evaluated on each input separately --
  followed by a pairwise cost that involves both inputs, as in:

  .. math::

    c(x,y) = norm(x) + norm(y) + pairwise(x,y)

  If the norm function is not implemented, that value is handled as a 0.
  """

  # no norm function created by default.
  norm: Optional[Callable[[jnp.ndarray], Union[float, jnp.ndarray]]] = None

  @abc.abstractmethod
  def pairwise(self, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    pass

  def barycenter(self, weights: jnp.ndarray, xs: jnp.ndarray) -> jnp.ndarray:
    """Barycentric projection.

    Args:
      weights: Weights of the points.
      xs: Points to project.

    Returns:
      The barycentric projection.
    """
    raise NotImplementedError("Barycenter is not yet implemented.")

  @classmethod
  def _padder(cls, dim: int) -> jnp.ndarray:
    """Create a padding vector for easier jitting.

    Args:
      dim: Dimensionality of the data.

    Returns:
      The padding vector.
    """
    return jnp.zeros((1, dim))

  def __call__(self, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    cost = self.pairwise(x, y)
    if self.norm is None:
      return cost
    return cost + self.norm(x) + self.norm(y)

  def all_pairs(self, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    """Compute matrix of all costs (including norms) for vectors in x / y.

    Args:
      x: [num_a, d] jnp.ndarray
      y: [num_b, d] jnp.ndarray
    Returns:
      [num_a, num_b] matrix of cost evaluations.
    """
    return jax.vmap(lambda x_: jax.vmap(lambda y_: self(x_, y_))(y))(x)

  def all_pairs_pairwise(self, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    """Compute matrix of all pairwise-costs (no norms) for vectors in x / y.

    Args:
      x: [num_a, d] jnp.ndarray
      y: [num_b, d] jnp.ndarray
    Returns:
      [num_a, num_b] matrix of pairwise cost evaluations.
    """
    return jax.vmap(lambda x_: jax.vmap(lambda y_: self.pairwise(x_, y_))(y))(x)

  def tree_flatten(self):
    return (), None

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    del aux_data
    return cls(*children)


@jax.tree_util.register_pytree_node_class
class TICost(CostFn):
  """A class for translation invariant (TI) costs.

  Such costs are defined using a function :math:`h`, mapping vectors to
  real-values, to be used as:

  .. math::

    c(x,y) = h(z), z := x-y.

  If that cost function is used to form an Entropic map using the
  :cite:`brenier:91` theorem, then the user should ensure :math:`h` is
  strictly convex, as well as provide the Legendre transform of :math:`h`,
  whose gradient is necessarily the inverse of the gradient of :math:`h`.
  """

  @abc.abstractmethod
  def h(self, z: jnp.ndarray) -> float:
    """TI function acting on difference of :math:`x-y` to output cost."""

  def h_legendre(self, z: jnp.ndarray) -> float:
    """Legendre transform of :func:`h` when it is convex."""
    raise NotImplementedError("`h_legendre` not implemented.")

  def pairwise(self, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    """Compute cost as evaluation of :func:`h` on :math:`x-y`."""
    return self.h(x - y)


@jax.tree_util.register_pytree_node_class
class SqPNorm(TICost):
  """Squared p-norm of the difference of two vectors.

  For details on the derivation of the Legendre transform of the norm, see e.g.
  the reference :cite:`boyd:04`, p.93/94.

  Args:
    p: Power of the p-norm.
  """

  def __init__(self, p: float):
    super().__init__()
    assert p >= 1.0, "p parameter in sq. p-norm should be >= 1.0"
    self.p = p
    self.q = 1. / (1. - 1. / self.p) if p > 1.0 else "inf"

  def h(self, z: jnp.ndarray) -> float:
    return 0.5 * jnp.linalg.norm(z, self.p) ** 2

  def h_legendre(self, z: jnp.ndarray) -> float:
    return 0.5 * jnp.linalg.norm(z, self.q) ** 2

  def tree_flatten(self):
    return (), (self.p,)

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    del children
    return cls(aux_data[0])


@jax.tree_util.register_pytree_node_class
class PNorm(TICost):
  """p-norm (to the power p) of the difference of two vectors.

  Args:
    p: Power of the p-norm.
  """

  def __init__(self, p: float):
    super().__init__()
    assert p >= 1.0, "p parameter in p-norm should be >= 1.0"
    self.p = p
    # TODO(marcocuturi): fix case when `p=1`
    self.q = 1. / (1. - 1. / self.p) if p > 1. else "inf"

  def h(self, z: jnp.ndarray) -> float:
    return jnp.linalg.norm(z, self.p) ** self.p / self.p

  def h_legendre(self, z: jnp.ndarray) -> float:
    return jnp.linalg.norm(z, self.q) ** self.q / self.q

  def tree_flatten(self):
    return (), (self.p,)

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    del children
    return cls(aux_data[0])


@jax.tree_util.register_pytree_node_class
class Euclidean(CostFn):
  """Euclidean distance.

  Note that the Euclidean distance is not cast as a
  :class:`~ott.geometry.costs.TICost`, since this would correspond to :math:`h`
  being :func:`jax.numpy.linalg.norm`, whose gradient is not invertible,
  because the function is not strictly convex (it is linear on rays).
  """

  def pairwise(self, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    """Compute Euclidean norm."""
    return jnp.linalg.norm(x - y)


@jax.tree_util.register_pytree_node_class
class SqEuclidean(TICost):
  """Squared Euclidean distance."""

  def norm(self, x: jnp.ndarray) -> Union[float, jnp.ndarray]:
    """Compute squared Euclidean norm for vector."""
    return jnp.sum(x ** 2, axis=-1)

  def pairwise(self, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    """Compute minus twice the dot-product between vectors."""
    return -2. * jnp.vdot(x, y)

  def h(self, z: jnp.ndarray) -> float:
    return jnp.sum(z ** 2)

  def h_legendre(self, z: jnp.ndarray) -> float:
    return 0.25 * jnp.sum(z ** 2)

  def barycenter(self, weights: jnp.ndarray, xs: jnp.ndarray) -> jnp.ndarray:
    """Output barycenter of vectors when using squared-Euclidean distance."""
    return jnp.average(xs, weights=weights, axis=0)


@jax.tree_util.register_pytree_node_class
class Cosine(CostFn):
  """Cosine distance cost function.

  Args:
    ridge: Ridge regularization.
  """

  def __init__(self, ridge: float = 1e-8):
    super().__init__()
    self._ridge = ridge

  def pairwise(self, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    """Cosine distance between vectors, denominator regularized with ridge."""
    ridge = self._ridge
    x_norm = jnp.linalg.norm(x, axis=-1)
    y_norm = jnp.linalg.norm(y, axis=-1)
    cosine_similarity = jnp.vdot(x, y) / (x_norm * y_norm + ridge)
    cosine_distance = 1.0 - cosine_similarity
    # similarity is in [-1, 1], clip because of numerical imprecisions
    return jnp.clip(cosine_distance, 0., 2.)

  @classmethod
  def _padder(cls, dim: int) -> jnp.ndarray:
    return jnp.ones((1, dim))


@jax.tree_util.register_pytree_node_class
class Bures(CostFn):
  """Bures distance between a pair of (mean, cov matrix) raveled as vectors.

  Args:
    dimension: Dimensionality of the data.
    kwargs: Keyword arguments for :func:`ott.math.matrix_square_root.sqrtm`.
  """

  def __init__(self, dimension: int, **kwargs: Any):
    super().__init__()
    self._dimension = dimension
    self._sqrtm_kw = kwargs

  def norm(self, x: jnp.ndarray) -> jnp.ndarray:
    """Compute norm of Gaussian, sq. 2-norm of mean + trace of covariance."""
    mean, cov = x_to_means_and_covs(x, self._dimension)
    norm = jnp.sum(mean ** 2, axis=-1)
    norm += jnp.trace(cov, axis1=-2, axis2=-1)
    return norm

  def pairwise(self, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    """Compute - 2 x Bures dot-product."""
    mean_x, cov_x = x_to_means_and_covs(x, self._dimension)
    mean_y, cov_y = x_to_means_and_covs(y, self._dimension)
    mean_dot_prod = jnp.vdot(mean_x, mean_y)
    sq_x = matrix_square_root.sqrtm(cov_x, self._dimension, **self._sqrtm_kw)[0]
    sq_x_y_sq_x = jnp.matmul(sq_x, jnp.matmul(cov_y, sq_x))
    sq__sq_x_y_sq_x = matrix_square_root.sqrtm(
        sq_x_y_sq_x, self._dimension, **self._sqrtm_kw
    )[0]
    return -2 * (mean_dot_prod + jnp.trace(sq__sq_x_y_sq_x, axis1=-2, axis2=-1))

  def covariance_fixpoint_iter(
      self,
      covs: jnp.ndarray,
      lambdas: jnp.ndarray,
      rtol: float = 1e-2
  ) -> jnp.ndarray:
    """Iterate fix-point updates to compute barycenter of Gaussians."""

    @functools.partial(jax.vmap, in_axes=[None, 0, 0])
    def scale_covariances(
        cov_sqrt: jnp.ndarray, cov_i: jnp.ndarray, lambda_i: jnp.ndarray
    ) -> jnp.ndarray:
      """Iterate update needed to compute barycenter of covariances."""
      return lambda_i * matrix_square_root.sqrtm_only(
          (cov_sqrt @ cov_i) @ cov_sqrt
      )

    def cond_fn(iteration: int, constants: Tuple[Any, ...], state) -> bool:
      del iteration, constants
      _, diff = state
      return diff > rtol

    def body_fn(
        iteration: int, constants: Tuple[Any, ...],
        state: Tuple[jnp.ndarray, float], compute_error: bool
    ) -> Tuple[jnp.ndarray, float]:
      del iteration, constants, compute_error
      cov, _ = state
      cov_sqrt, cov_inv_sqrt, _ = matrix_square_root.sqrtm(cov)
      scaled_cov = jnp.linalg.matrix_power(
          jnp.sum(scale_covariances(cov_sqrt, covs, lambdas), axis=0), 2
      )
      next_cov = (cov_inv_sqrt @ scaled_cov) @ cov_inv_sqrt
      diff = jnp.sum((next_cov - cov) ** 2) / jnp.prod(jnp.array(cov.shape))
      return next_cov, diff

    def init_state() -> Tuple[jnp.ndarray, float]:
      cov_init = jnp.eye(self._dimension)
      diff = jnp.inf
      return cov_init, diff

    cov, _ = fixed_point_loop.fixpoint_iter(
        cond_fn=cond_fn,
        body_fn=body_fn,
        min_iterations=10,
        max_iterations=500,
        inner_iterations=1,
        constants=(),
        state=init_state()
    )
    return cov

  def barycenter(self, weights: jnp.ndarray, xs: jnp.ndarray) -> jnp.ndarray:
    """Compute the Bures barycenter of weighted Gaussian distributions.

    Implements the fixed point approach proposed in :cite:`alvarez-esteban:16`
    for the computation of the mean and the covariance of the barycenter of
    weighted Gaussian distributions.

    Args:
      weights: The barycentric weights.
      xs: The points to be used in the computation of the barycenter, where
        each point is described by a concatenation of the mean and the
        covariance (raveled).

    Returns:
      A concatenation of the mean and the raveled covariance of the barycenter.
    """
    # Ensure that barycentric weights sum to 1.
    weights = weights / jnp.sum(weights)
    mus, covs = x_to_means_and_covs(xs, self._dimension)
    mu_bary = jnp.sum(weights[:, None] * mus, axis=0)
    cov_bary = self.covariance_fixpoint_iter(covs=covs, lambdas=weights)
    barycenter = mean_and_cov_to_x(mu_bary, cov_bary, self._dimension)
    return barycenter

  @classmethod
  def _padder(cls, dim: int) -> jnp.ndarray:
    """Pad with concatenated zero means and \
      raveled identity covariance matrix."""
    dimension = int((-1 + math.sqrt(1 + 4 * dim)) / 2)
    padding = mean_and_cov_to_x(
        jnp.zeros((dimension,)), jnp.eye(dimension), dimension
    )
    return padding[jnp.newaxis, :]

  def tree_flatten(self):
    return (), (self._dimension, self._sqrtm_kw)

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    del children
    return cls(aux_data[0], **aux_data[1])


@jax.tree_util.register_pytree_node_class
class UnbalancedBures(CostFn):
  """Unbalanced Bures distance between two triplets of `(mass, mean, cov)`.

  This cost uses the notation defined in :cite:`janati:20`, eq. 37, 39, 40.

  Args:
    dimension: Dimensionality of the data.
    sigma: Entropic regularization.
    gamma: KL-divergence regularization for the marginals.
    kwargs: Keyword arguments for :func:`~ott.math.matrix_square_root.sqrtm`.
  """

  def __init__(
      self,
      dimension: int,
      *,
      sigma: float = 1.0,
      gamma: float = 1.0,
      **kwargs: Any,
  ):
    super().__init__()
    self._dimension = dimension
    self._sigma = sigma
    self._gamma = gamma
    self._sqrtm_kw = kwargs

  def norm(self, x: jnp.ndarray) -> jnp.ndarray:
    """Compute norm of Gaussian for unbalanced Bures.

    Args:
      x: Array of shape ``[n_points + n_points + n_dim ** 2,]``, potentially
        batched, corresponding to the raveled mass, means and the covariance
        matrix.

    Returns:
      The norm, array of shape ``[]`` or ``[batch,]`` in the batched case.
    """
    return self._gamma * x[..., 0]

  def pairwise(self, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    """Compute dot-product for unbalanced Bures.

    Args:
      x: Array of shape ``[n_points + n_points + n_dim ** 2,]``
        corresponding to the raveled mass, means and the covariance matrix.
      y: Array of shape ``[n_points + n_points + n_dim ** 2,]``
        corresponding to the raveled mass, means and the covariance matrix.

    Returns:
      The cost.
    """
    # Sets a few constants
    gam = self._gamma
    sig2 = self._sigma ** 2
    lam = sig2 + gam / 2.0
    tau = gam / (2.0 * lam)

    # Extracts mass, mean vector, covariance matrices
    mass_x, mass_y = x[0], y[0]
    mean_x, cov_x = x_to_means_and_covs(x[1:], self._dimension)
    mean_y, cov_y = x_to_means_and_covs(y[1:], self._dimension)

    diff_means = mean_x - mean_y

    # Identity matrix of suitable size
    iden = jnp.eye(self._dimension, dtype=x.dtype)

    # Creates matrices needed in the computation
    tilde_a = 0.5 * gam * (iden - lam * jnp.linalg.inv(cov_x + lam * iden))
    tilde_b = 0.5 * gam * (iden - lam * jnp.linalg.inv(cov_y + lam * iden))

    tilde_a_b = jnp.matmul(tilde_a, tilde_b)
    c_mat = matrix_square_root.sqrtm(
        1 / tau * tilde_a_b + 0.25 * (sig2 ** 2) * iden, **self._sqrtm_kw
    )[0]
    c_mat -= 0.5 * sig2 * iden

    # Computes log determinants (their sign should be >0).
    sldet_c, ldet_c = jnp.linalg.slogdet(c_mat)
    sldet_t_ab, ldet_t_ab = jnp.linalg.slogdet(tilde_a_b)
    sldet_ab, ldet_ab = jnp.linalg.slogdet(jnp.matmul(cov_x, cov_y))
    sldet_c_ab, ldet_c_ab = jnp.linalg.slogdet(c_mat - 2.0 * tilde_a_b / gam)

    # Gathers all these results to compute log total mass of transport
    log_m_pi = (0.5 * self._dimension * sig2 / (gam + sig2)) * jnp.log(sig2)
    log_m_pi += (1.0 / (tau + 1.0)) * (
        jnp.log(mass_x) + jnp.log(mass_y) + ldet_c + 0.5 *
        (tau * ldet_t_ab - ldet_ab)
    )
    log_m_pi += -jnp.sum(
        diff_means * jnp.linalg.solve(cov_x + cov_y + lam * iden, diff_means)
    ) / (2.0 * (tau + 1.0))
    log_m_pi += -0.5 * ldet_c_ab

    # if all logdet signs are 1, output value, nan otherwise
    pos_signs = (sldet_c + sldet_c_ab + sldet_t_ab + sldet_t_ab) == 4

    return jax.lax.cond(
        pos_signs, lambda: 2 * sig2 * mass_x * mass_y - 2 *
        (sig2 + gam) * jnp.exp(log_m_pi), lambda: jnp.nan
    )

  def tree_flatten(self):
    return (), (self._dimension, self._sigma, self._gamma, self._sqrtm_kw)

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    del children
    dim, sigma, gamma, kwargs = aux_data
    return cls(dim, sigma=sigma, gamma=gamma, **kwargs)


def x_to_means_and_covs(x: jnp.ndarray,
                        dimension: int) -> Tuple[jnp.ndarray, jnp.ndarray]:
  """Extract means and covariance matrices of Gaussians from raveled vector.

  Args:
    x: [num_gaussians, dimension, (1 + dimension)] array of concatenated means
      and covariances (raveled) dimension: the dimension of the Gaussians.

  Returns:
    means: [num_gaussians, dimension] array that holds the means.
    covariances: [num_gaussians, dimension] array that holds the covariances.
  """
  x = jnp.atleast_2d(x)
  means = x[:, :dimension]
  covariances = jnp.reshape(
      x[:, dimension:dimension + dimension ** 2], (-1, dimension, dimension)
  )
  return jnp.squeeze(means), jnp.squeeze(covariances)


def mean_and_cov_to_x(
    mean: jnp.ndarray, covariance: jnp.ndarray, dimension: int
) -> jnp.ndarray:
  """Ravel a Gaussian's mean and covariance matrix to d(1 + d) vector."""
  x = jnp.concatenate((mean, jnp.reshape(covariance, (dimension * dimension))))
  return x



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python 3
"""Pytree for a Gaussian mixture model."""

from typing import List, Optional, Tuple, Union

import jax
import jax.numpy as jnp

from ott.tools.gaussian_mixture import gaussian, linalg, probabilities, scale_tril


def get_summary_stats_from_points_and_assignment_probs(
    points: jnp.ndarray, point_weights: jnp.ndarray,
    assignment_probs: jnp.ndarray
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
  """Get component summary stats from points and component probabilities.

  Args:
    points: array of points, shape (n, n_dim)
    point_weights: array of weights for the points, shape (n,)
    assignment_probs: array of component assignment probabilities for the
      points, shape (n, n_components)

  Returns:
    Tuple containing for each component,
    * the sample mean for each component, shape (n_components, n_dim)
    * the sample covariance for each component,
        shape (n_components, n_dim, n_dim)
    * the weight for each component,
        shape (n_components,)
  """

  def component_from_points(points, point_weights, assignment_probs):
    component_weight = (
        jnp.sum(point_weights * assignment_probs) / jnp.sum(point_weights)
    )
    component_mean, component_cov = linalg.get_mean_and_cov(
        points=points, weights=point_weights * assignment_probs
    )
    return component_mean, component_cov, component_weight

  components_from_points_fn = jax.vmap(
      component_from_points, in_axes=(None, None, 1), out_axes=0
  )

  return components_from_points_fn(points, point_weights, assignment_probs)


@jax.tree_util.register_pytree_node_class
class GaussianMixture:
  """Pytree for a Gaussian Mixture model."""

  def __init__(
      self, loc: jnp.ndarray, scale_params: jnp.ndarray,
      component_weight_ob: probabilities.Probabilities
  ):
    self._loc = loc
    self._scale_params = scale_params
    self._component_weight_ob = component_weight_ob

  @classmethod
  def from_random(
      cls,
      key: jnp.ndarray,
      n_components: int,
      n_dimensions: int,
      stdev_mean: float = 0.1,
      stdev_cov: float = 0.1,
      stdev_weights: float = 0.1,
      ridge: Union[float, jnp.array] = 0,
      dtype: Optional[jnp.dtype] = None
  ) -> 'GaussianMixture':
    """Construct a random GMM."""
    loc = []
    scale_params = []
    for _ in range(n_components):
      key, subkey = jax.random.split(key)
      component = gaussian.Gaussian.from_random(
          key=subkey,
          n_dimensions=n_dimensions,
          stdev_mean=stdev_mean,
          stdev_cov=stdev_cov,
          ridge=ridge,
          dtype=dtype
      )
      loc.append(component.loc)
      scale_params.append(component.scale.params)
    loc = jnp.stack(loc, axis=0)
    scale_params = jnp.stack(scale_params, axis=0)
    weight_ob = probabilities.Probabilities.from_random(
        key=subkey, n_dimensions=n_components, stdev=stdev_weights, dtype=dtype
    )
    return cls(
        loc=loc, scale_params=scale_params, component_weight_ob=weight_ob
    )

  @classmethod
  def from_mean_cov_component_weights(
      cls, mean: jnp.ndarray, cov: jnp.ndarray, component_weights: jnp.ndarray
  ):
    """Construct a GMM from means, covariances, and component weights."""
    scale_params = []
    for i in range(cov.shape[0]):
      scale_params.append(scale_tril.ScaleTriL.from_covariance(cov[i]).params)
    scale_params = jnp.stack(scale_params, axis=0)
    weight_ob = probabilities.Probabilities.from_probs(component_weights)
    return cls(
        loc=mean, scale_params=scale_params, component_weight_ob=weight_ob
    )

  @classmethod
  def from_points_and_assignment_probs(
      cls,
      points: jnp.ndarray,
      point_weights: jnp.ndarray,
      assignment_probs: jnp.ndarray,
  ) -> 'GaussianMixture':
    """Estimate a GMM from points and a set of component probabilities."""
    mean, cov, wts = get_summary_stats_from_points_and_assignment_probs(
        points=points,
        point_weights=point_weights,
        assignment_probs=assignment_probs
    )
    return cls.from_mean_cov_component_weights(
        mean=mean, cov=cov, component_weights=wts
    )

  @property
  def dtype(self):
    return self.loc.dtype

  @property
  def n_dimensions(self):
    return self._loc.shape[-1]

  @property
  def n_components(self):
    return self._loc.shape[-2]

  @property
  def loc(self) -> jnp.ndarray:
    return self._loc

  @property
  def scale_params(self) -> jnp.ndarray:
    return self._scale_params

  @property
  def cholesky(self) -> jnp.ndarray:
    size = self.n_dimensions

    def _get_cholesky(scale_params):
      return scale_tril.ScaleTriL(params=scale_params, size=size).cholesky()

    return jax.vmap(_get_cholesky, in_axes=0, out_axes=0)(self.scale_params)

  @property
  def covariance(self) -> jnp.ndarray:
    size = self.n_dimensions

    def _get_covariance(scale_params):
      return scale_tril.ScaleTriL(params=scale_params, size=size).covariance()

    return jax.vmap(_get_covariance, in_axes=0, out_axes=0)(self.scale_params)

  @property
  def component_weight_ob(self) -> probabilities.Probabilities:
    return self._component_weight_ob

  @property
  def component_weights(self) -> jnp.ndarray:
    return self._component_weight_ob.probs()

  def log_component_weights(self) -> jnp.ndarray:
    return self._component_weight_ob.log_probs()

  def _get_normal(
      self, loc: jnp.ndarray, scale_params: jnp.ndarray
  ) -> gaussian.Gaussian:
    size = loc.shape[-1]
    return gaussian.Gaussian(
        loc=loc, scale=scale_tril.ScaleTriL(params=scale_params, size=size)
    )

  def get_component(self, index: int) -> gaussian.Gaussian:
    """Get the specified GMM component."""
    return self._get_normal(
        loc=self.loc[index], scale_params=self.scale_params[index]
    )

  def components(self) -> List[gaussian.Gaussian]:
    """Get a list of all GMM components."""
    return [self.get_component(i) for i in range(self.n_components)]

  def sample(self, key: jnp.ndarray, size: int) -> jnp.ndarray:
    """Generate samples from the distribution."""
    subkey0, subkey1 = jax.random.split(key)
    component = self.component_weight_ob.sample(key=subkey0, size=size)
    std_samples = jax.random.normal(
        key=subkey1, shape=(size, self.n_dimensions)
    )

    def _transform_single_component(k, scale, loc):

      def _transform_single_value(single_component, single_x):
        return jax.lax.cond(
            single_component == k,
            lambda x: jnp.matmul(scale, x[:, None])[:, 0] + loc, jnp.zeros_like,
            single_x
        )

      return jax.vmap(_transform_single_value)(component, std_samples)

    return jnp.sum(
        jax.vmap(_transform_single_component)
        (jnp.arange(self.n_components), self.cholesky, self.loc),
        axis=0
    )

  def conditional_log_prob(self, x: jnp.ndarray) -> jnp.ndarray:
    """Compute the component-conditional log probability of x.

    Args:
      x: (n, n_dimensions) array of points

    Returns:
      (n, n_components) array of the log probability of x conditioned on it
      having come from each component.
    """

    def _log_prob_single_component(
        loc: jnp.ndarray, scale_params: jnp.ndarray, x: jnp.ndarray
    ):
      norm = self._get_normal(loc=loc, scale_params=scale_params)
      return norm.log_prob(x)

    conditional_log_prob_fn = jax.vmap(
        _log_prob_single_component, in_axes=(0, 0, None), out_axes=1
    )
    return conditional_log_prob_fn(self._loc, self._scale_params, x)

  def log_prob(self, x: jnp.ndarray) -> jnp.ndarray:
    """Compute the log probability of the observations x.

    Args:
      x: (n, n_dimensions) array of points

    Returns:
      (n,) array of log probabilities.
    """
    # p(x) = \sum_i p(x|c_i) p(c_i)
    log_prob_conditional = self.conditional_log_prob(x)
    log_component_weight = self.log_component_weights()
    return jax.scipy.special.logsumexp(
        log_prob_conditional + log_component_weight[None, :], axis=-1
    )

  def get_log_component_posterior(self, x: jnp.ndarray) -> jnp.ndarray:
    """Compute the posterior probability that x came from each component.

    Args:
      x: (n, n_dimensions) array of points

    Returns:
      (n, n_components) array of poster component log probabilities.
    """
    # p(x | c_i) = p(x, c_i) / p(c_i) => p(x, c_i) = p(x | c_i) p(c_i)
    # p(c_i | x) = p(x, c_i) / p(x)
    #            = p(x | c_i) p(c_i) / sum_j(p(x | c_j)p(c_j))
    log_prob_conditional = self.conditional_log_prob(x)
    log_component_weight = self.log_component_weights()
    log_prob_unnorm = log_prob_conditional + log_component_weight[None, :]
    return log_prob_unnorm - jax.scipy.special.logsumexp(
        log_prob_unnorm, axis=-1, keepdims=True
    )

  def has_nans(self) -> bool:
    for leaf in jax.tree_util.tree_leaves(self):
      if jnp.any(~jnp.isfinite(leaf)):
        return True
    return False

  def tree_flatten(self):
    children = (self.loc, self.scale_params, self.component_weight_ob)
    aux_data = {}
    return children, aux_data

  @classmethod
  def tree_unflatten(cls, aux_data, children):
    return cls(*children, **aux_data)

  def __repr__(self):
    class_name = type(self).__name__
    children, aux = self.tree_flatten()
    return '{}({})'.format(
        class_name, ', '.join([repr(c) for c in children] +
                              [f'{k}: {repr(v)}' for k, v in aux.items()])
    )

  def __hash__(self):
    return jax.tree_util.tree_flatten(self).__hash__()

  def __eq__(self, other):
    return jax.tree_util.tree_flatten(self) == jax.tree_util.tree_flatten(other)



from functools import partial
from typing import Any, Dict, NamedTuple, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp

from ott.geometry import pointcloud
from ott.math import fixed_point_loop
from ott.problems.linear import linear_problem
from ott.problems.quadratic import gw_barycenter
from ott.solvers import was_solver
from ott.solvers.quadratic import gromov_wasserstein

__all__ = ["GWBarycenterState", "GromovWassersteinBarycenter"]


class GWBarycenterState(NamedTuple):
  """Holds the state of the \
  :class:`~ott.problems.quadratic.gw_barycenter.GWBarycenterProblem`.

  Args:
    c: Barycenter cost matrix of shape ``[bar_size, bar_size]``.
    x: Barycenter features of shape ``[bar_size, ndim_fused]``.
      Only used in the fused case.
    a: Weights of the barycenter of shape ``[bar_size,]``.
    errors: Array of shape
      ``[max_iter, num_measures, quad_max_iter, lin_outer_iter]`` containing
      the GW errors at each iteration.
    costs: Array of shape ``[max_iter,]`` containing the cost at each iteration.
    gw_convergence: Array of shape ``[max_iter,]`` containing the convergence
      of all GW problems at each iteration.
  """
  cost: Optional[jnp.ndarray] = None
  x: Optional[jnp.ndarray] = None
  a: Optional[jnp.ndarray] = None
  errors: Optional[jnp.ndarray] = None
  costs: Optional[jnp.ndarray] = None
  gw_convergence: Optional[jnp.ndarray] = None

  def set(self, **kwargs: Any) -> 'GWBarycenterState':
    """Return a copy of self, possibly with overwrites."""
    return self._replace(**kwargs)


@jax.tree_util.register_pytree_node_class
class GromovWassersteinBarycenter(was_solver.WassersteinSolver):
  """Gromov-Wasserstein barycenter solver of the \
  :class:`~ott.problems.quadratic.gw_barycenter.GWBarycenterProblem`.

  Args:
    epsilon: Entropy regulariser.
    min_iterations: Minimum number of iterations.
    max_iterations: Maximum number of outermost iterations.
    threshold: Convergence threshold.
    jit: Whether to jit the iteration loop.
    store_inner_errors: Whether to store the errors of the GW solver, as well
      as its linear solver, at each iteration for each measure.
    quad_solver: The GW solver.
    kwargs: Keyword argument for
      :class:`~ott.solvers.quadratic.gromov_wasserstein.GromovWasserstein`.
      Only used when ``quad_solver = None``.
  """

  def __init__(
      self,
      epsilon: Optional[float] = None,
      min_iterations: int = 5,
      max_iterations: int = 50,
      threshold: float = 1e-3,
      jit: bool = True,
      store_inner_errors: bool = False,
      quad_solver: Optional[gromov_wasserstein.GromovWasserstein] = None,
      # TODO(michalk8): maintain the API compatibility with `was_solver`
      # but makes passing kwargs with the same name to `quad_solver` impossible
      # will be fixed when refactoring the solvers
      # note that `was_solver` also suffers from this
      **kwargs: Any,
  ):
    super().__init__(
        epsilon=epsilon,
        min_iterations=min_iterations,
        max_iterations=max_iterations,
        threshold=threshold,
        jit=jit,
        store_inner_errors=store_inner_errors
    )
    self._quad_solver = quad_solver
    if quad_solver is None:
      kwargs["epsilon"] = epsilon
      # TODO(michalk8): store only GW errors?
      kwargs["store_inner_errors"] = store_inner_errors
      self._quad_solver = gromov_wasserstein.GromovWasserstein(**kwargs)

  def __call__(
      self, problem: gw_barycenter.GWBarycenterProblem, bar_size: int,
      **kwargs: Any
  ) -> GWBarycenterState:
    """Solver the (fused) GW barycenter problem.

    Args:
      problem: The GW barycenter problem.
      bar_size: Size of the barycenter.
      kwargs: Keyword arguments for :meth:`init_state`.

    Returns:
      The solution.
    """
    bar_fn = jax.jit(iterations, static_argnums=1) if self.jit else iterations
    state = self.init_state(problem, bar_size, **kwargs)
    state = bar_fn(solver=self, problem=problem, init_state=state)
    return self.output_from_state(state)

  def init_state(
      self,
      problem: gw_barycenter.GWBarycenterProblem,
      bar_size: int,
      bar_init: Optional[Union[jnp.ndarray, Tuple[jnp.ndarray,
                                                  jnp.ndarray]]] = None,
      a: Optional[jnp.ndarray] = None,
      seed: int = 0,
  ) -> GWBarycenterState:
    """Initialize the (fused) Gromov-Wasserstein barycenter state.

    Args:
      problem: The barycenter problem.
      bar_size: Size of the barycenter.
      bar_init: Initial barycenter value. Can be one of the following:

        - ``None`` - randomly initialize the barycenter.
        - :class:`jax.numpy.ndarray` - barycenter cost matrix of shape
          ``[bar_size, bar_size]``.
          Only used in the non-fused case.
        - :class:`tuple` of :class:`jax.numpy.ndarray` - the 1st array
          corresponds to a cost matrix of shape ``[bar_size, bar_size]``,
          the 2nd array is a ``[bar_size, ndim_fused]`` feature matrix used in
          the fused case.

      a: An array of shape ``[bar_size,]`` containing the barycenter weights.
      seed: Random seed used when ``bar_init = None``.

    Returns:
      The initial barycenter state.
    """
    if a is None:
      a = jnp.ones((bar_size,)) / bar_size
    else:
      assert a.shape == (bar_size,)

    if bar_init is None:
      _, b = problem.segmented_y_b
      rng = jax.random.PRNGKey(seed)
      keys = jax.random.split(rng, problem.num_measures)
      linear_solver = self._quad_solver.linear_ot_solver

      transports = init_transports(linear_solver, keys, a, b, problem.epsilon)
      x = problem.update_features(transports, a) if problem.is_fused else None
      cost = problem.update_barycenter(transports, a)
    else:
      cost, x = bar_init if isinstance(bar_init, tuple) else (bar_init, None)
      assert cost.shape == (bar_size, bar_size)
      if problem.is_fused:
        assert x is not None, "Barycenter features are not initialized."
        assert x.shape == (bar_size, problem.ndim_fused)

    num_iter = self.max_iterations
    if self.store_inner_errors:
      # TODO(michalk8): in the future, think about how to do this in general
      errors = -jnp.ones((
          num_iter, problem.num_measures, self._quad_solver.max_iterations,
          self._quad_solver.linear_ot_solver.outer_iterations
      ))
    else:
      errors = None

    costs = -jnp.ones((num_iter,))
    gw_convergence = -jnp.ones((num_iter,))
    return GWBarycenterState(
        cost=cost,
        x=x,
        a=a,
        errors=errors,
        costs=costs,
        gw_convergence=gw_convergence
    )

  def update_state(
      self,
      state: GWBarycenterState,
      iteration: int,
      problem: gw_barycenter.GWBarycenterProblem,
      store_errors: bool = True,
  ) -> Tuple[float, bool, jnp.ndarray, Optional[jnp.ndarray]]:

    def solve_gw(
        state: GWBarycenterState, b: jnp.ndarray, y: jnp.ndarray,
        f: Optional[jnp.ndarray]
    ) -> Tuple[float, bool, jnp.ndarray, Optional[jnp.ndarray]]:
      quad_problem = problem._create_problem(state, y=y, b=b, f=f)
      out = self._quad_solver(quad_problem)
      return (
          out.reg_gw_cost, out.converged, out.matrix,
          out.errors if store_errors else None
      )

    in_axes = [None, 0, 0]
    in_axes += [0] if problem.is_fused else [None]
    solve_fn = jax.vmap(solve_gw, in_axes=in_axes)

    y, b = problem.segmented_y_b
    y_f = problem.segmented_y_fused
    costs, convergeds, transports, errors = solve_fn(state, b, y, y_f)

    cost = jnp.sum(costs * problem.weights)
    costs = state.costs.at[iteration].set(cost)
    converged = jnp.all(convergeds)
    gw_convergence = state.gw_convergence.at[iteration].set(converged)

    if self.store_inner_errors:
      errors = state.errors.at[iteration, ...].set(errors)
    else:
      errors = None

    x = problem.update_features(
        transports, state.a
    ) if problem.is_fused else state.x
    cost = problem.update_barycenter(transports, state.a)
    return state.set(
        cost=cost,
        x=x,
        costs=costs,
        errors=errors,
        gw_convergence=gw_convergence
    )

  def output_from_state(self, state: GWBarycenterState) -> GWBarycenterState:
    """No-op."""
    # TODO(michalk8): just for consistency with continuous barycenter
    # will be refactored in the future to create an output
    return state

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:
    children, aux = super().tree_flatten()
    return children + [self._quad_solver], aux

  @classmethod
  def tree_unflatten(
      cls, aux_data: Dict[str, Any], children: Sequence[Any]
  ) -> "GromovWassersteinBarycenter":
    epsilon, _, threshold, quad_solver = children
    return cls(
        epsilon=epsilon,
        threshold=threshold,
        quad_solver=quad_solver,
        **aux_data,
    )


@partial(jax.vmap, in_axes=[None, 0, None, 0, None])
def init_transports(
    solver, key: jnp.ndarray, a: jnp.ndarray, b: jnp.ndarray,
    epsilon: Optional[float]
) -> jnp.ndarray:
  """Initialize random 2D point cloud and solve the linear OT problem.

  Args:
    solver: Linear OT solver.
    key: Random key.
    a: Source marginals (e.g., for barycenter) of shape ``[bar_size,]``.
    b: Target marginals of shape ``[max_measure_size,]``.
    epsilon: Entropy regularization.

  Returns:
    Transport map of shape ``[bar_size, max_measure_size]``.
  """
  key1, key2 = jax.random.split(key, 2)
  x = jax.random.normal(key1, shape=(len(a), 2))
  y = jax.random.normal(key2, shape=(len(b), 2))
  geom = pointcloud.PointCloud(
      x, y, epsilon=epsilon, src_mask=a > 0, tgt_mask=b > 0
  )
  problem = linear_problem.LinearProblem(geom, a=a, b=b)
  return solver(problem).matrix


def iterations(
    solver: GromovWassersteinBarycenter,
    problem: gw_barycenter.GWBarycenterProblem, init_state: GWBarycenterState
) -> GWBarycenterState:

  def cond_fn(
      iteration: int, constants: GromovWassersteinBarycenter,
      state: GWBarycenterState
  ) -> bool:
    solver, _ = constants
    return solver._continue(state, iteration)

  def body_fn(
      iteration, constants: Tuple[GromovWassersteinBarycenter,
                                  gw_barycenter.GWBarycenterProblem],
      state: GWBarycenterState, compute_error: bool
  ) -> GWBarycenterState:
    del compute_error  # always assumed true
    solver, problem = constants
    return solver.update_state(state, iteration, problem)

  state = fixed_point_loop.fixpoint_iter(
      cond_fn=cond_fn,
      body_fn=body_fn,
      min_iterations=solver.min_iterations,
      max_iterations=solver.max_iterations,
      inner_iterations=1,
      constants=(solver, problem),
      state=init_state,
  )
  return state



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Losses for the fairness example."""

import functools

import jax
import jax.numpy as jnp

import ott


def binary_cross_entropy(logits, labels):
  return jnp.sum(-labels * jnp.log(logits) - (1 - labels) * jnp.log(1 - logits))


def compute_metrics(logits, labels):
  loss = binary_cross_entropy(logits, labels)
  accuracy = jnp.mean((logits > 0.5) == labels)
  metrics = {
      'loss': loss,
      'accuracy': accuracy,
  }
  metrics = jax.lax.pmean(metrics, axis_name='batch')
  return metrics


@functools.partial(jax.jit, static_argnums=(2, 3))
def sort_group(
    inputs: jnp.ndarray, in_group: jnp.ndarray, quantization: int,
    epsilon: float
) -> jnp.ndarray:
  """Sorts and quantizes only the member of the given group.

  Args:
    inputs: 1D array to be sorted.
    in_group: a 1D array of 0s and 1s indicating if the element is part of the
      group or not.
    quantization: the number of values the sorted values output should be mapped
      onto.
    epsilon: sinkhorn entropic regularization.

  Returns:
    A sorted array of size `quantization`.
  """
  a = in_group / jnp.sum(in_group)
  b = jnp.ones(quantization) / quantization
  ot = ott.tools.soft_sort.transport_for_sort(inputs, a, b, epsilon=epsilon)
  return 1.0 / b * ot.apply(inputs, axis=0)


def fairness_regularizer(
    inputs: jnp.ndarray,
    groups: jnp.ndarray,
    quantization: int = 16,
    epsilon: float = 1e-2,
    num_groups: int = 2
):
  """Approximation of the wasserstein between the per-group distributions."""
  quantiles = jnp.stack([
      sort_group(inputs, groups == g, quantization, epsilon)
      for g in range(num_groups)
  ])
  weights = jnp.stack([jnp.sum(groups == g) for g in range(num_groups)]
                     ) / groups.shape[0]  # noqa: E124
  mean_quantile = jnp.sum(weights[:, None] * quantiles, axis=0)
  delta = jnp.where(
      quantiles, quantiles - mean_quantile, jnp.zeros_like(mean_quantile)
  )
  return jnp.mean(delta ** 2)



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Implementation of :cite:`amos:17` input convex neural networks (ICNN)."""

from typing import Any, Callable, Sequence, Tuple, Union

import jax
import jax.numpy as jnp
import optax
from flax import linen as nn
from flax.training import train_state
from jax.nn import initializers

from ott.math import matrix_square_root
from ott.solvers.nn.layers import PosDefPotentials, PositiveDense

__all__ = ["ICNN"]

PRNGKey = Any
Shape = Tuple[int]
Dtype = Any
Array = Any


class ICNN(nn.Module):
  """Input convex neural network (ICNN) architecture with initialization.

  Implementation of input convex neural networks as introduced in
  :cite:`amos:17` with initialization schemes proposed by :cite:`bunne:22`.

  Args:
    dim_hidden: sequence specifying size of hidden dimensions. The
      output dimension of the last layer is 1 by default.
    init_std: value of standard deviation of weight initialization method.
    init_fn: choice of initialization method for weight matrices (default:
      `jax.nn.initializers.normal`).
    act_fn: choice of activation function used in network architecture
      (needs to be convex, default: `nn.relu`).
    pos_weights: choice to enforce positivity of weight or use regularizer.
    dim_data: data dimensionality (default: 2).
    gaussian_map: data inputs of source and target measures for
      initialization scheme based on Gaussian approximation of input and
      target measure (if None, identity initialization is used).
  """

  dim_hidden: Sequence[int]
  init_std: float = 1e-1
  init_fn: Callable = jax.nn.initializers.normal
  act_fn: Callable = nn.relu
  pos_weights: bool = True
  dim_data: int = 2
  gaussian_map: Tuple[jnp.ndarray, jnp.ndarray] = None

  def setup(self) -> None:
    self.num_hidden = len(self.dim_hidden)

    if self.pos_weights:
      hid_dense = PositiveDense
      # this function needs to be the inverse map of function
      # used in PositiveDense layers
      rescale = hid_dense.inv_rectifier_fn
    else:
      hid_dense = nn.Dense
      rescale = lambda x: x
    self.use_init = False
    # check if Gaussian map was provided
    if self.gaussian_map is not None:
      factor, mean = self._compute_gaussian_map(self.gaussian_map)
    else:
      factor, mean = self._compute_identity_map(self.dim_data)

    w_zs = []
    # keep track of previous size to normalize accordingly
    normalization = 1
    # subsequent layers propagate value of potential provided by
    # first layer in x normalization factor is rescaled accordingly
    for i in range(self.num_hidden):
      w_zs.append(
          hid_dense(
              self.dim_hidden[i],
              kernel_init=initializers.constant(rescale(1.0 / normalization)),
              use_bias=False,
          )
      )
      normalization = self.dim_hidden[i]
    # final layer computes average, still with normalized rescaling
    w_zs.append(
        hid_dense(
            1,
            kernel_init=initializers.constant(rescale(1.0 / normalization)),
            use_bias=False,
        )
    )
    self.w_zs = w_zs

    w_xs = []
    # first square layer, initialized to identity
    w_xs.append(
        PosDefPotentials(
            self.dim_data,
            num_potentials=1,
            kernel_init=lambda *args, **kwargs: factor,
            bias_init=lambda *args, **kwargs: mean,
            use_bias=True,
        )
    )

    # subsequent layers reinjected into convex functions
    for i in range(self.num_hidden):
      w_xs.append(
          nn.Dense(
              self.dim_hidden[i],
              kernel_init=self.init_fn(self.init_std),
              bias_init=self.init_fn(self.init_std),
              use_bias=True,
          )
      )
    # final layer, to output number
    w_xs.append(
        nn.Dense(
            1,
            kernel_init=self.init_fn(self.init_std),
            bias_init=self.init_fn(self.init_std),
            use_bias=True,
        )
    )
    self.w_xs = w_xs

  @staticmethod
  def _compute_gaussian_map(
      inputs: Tuple[jnp.ndarray, jnp.ndarray]
  ) -> Tuple[jnp.ndarray, jnp.ndarray]:

    def compute_moments(x, reg=1e-4, sqrt_inv=False):
      shape = x.shape
      z = x.reshape(shape[0], -1)
      mu = jnp.expand_dims(jnp.mean(z, axis=0), 0)
      z = z - mu
      matmul = lambda a, b: jnp.matmul(a, b)
      sigma = jax.vmap(matmul)(jnp.expand_dims(z, 2), jnp.expand_dims(z, 1))
      # unbiased estimate
      sigma = jnp.sum(sigma, axis=0) / (shape[0] - 1)
      # regularize
      sigma = sigma + reg * jnp.eye(shape[1])

      if sqrt_inv:
        sigma_sqrt, sigma_inv_sqrt, _ = matrix_square_root.sqrtm(sigma)
        return sigma, sigma_sqrt, sigma_inv_sqrt, mu
      else:
        return sigma, mu

    source, target = inputs
    _, covs_sqrt, covs_inv_sqrt, mus = compute_moments(source, sqrt_inv=True)
    covt, mut = compute_moments(target, sqrt_inv=False)

    mo = matrix_square_root.sqrtm_only(
        jnp.dot(jnp.dot(covs_sqrt, covt), covs_sqrt)
    )
    A = jnp.dot(jnp.dot(covs_inv_sqrt, mo), covs_inv_sqrt)
    b = jnp.squeeze(mus) - jnp.linalg.solve(A, jnp.squeeze(mut))
    A = matrix_square_root.sqrtm_only(A)

    return jnp.expand_dims(A, 0), jnp.expand_dims(b, 0)

  @staticmethod
  def _compute_identity_map(input_dim: int) -> Tuple[jnp.ndarray, jnp.ndarray]:
    A = jnp.eye(input_dim).reshape((1, input_dim, input_dim))
    b = jnp.zeros((1, input_dim))

    return A, b

  @nn.compact
  def __call__(self, x: jnp.ndarray) -> float:
    for i in range(self.num_hidden + 2):
      if i == 0:
        z = self.w_xs[i](x)
      # apply both transform on hidden state and x
      # x is one step ahead as there is one more hidden layer for x
      else:
        z = jnp.add(self.w_zs[i - 1](z), self.w_xs[i](x))
      if i != 0 or i != self.num_hidden + 1:
        z = self.act_fn(z)
    return jnp.squeeze(z)

  def create_train_state(
      self,
      rng: jnp.ndarray,
      optimizer: optax.OptState,
      input: Union[int, Tuple[int, ...]],
  ) -> train_state.TrainState:
    """Create initial `TrainState`."""
    params = self.init(rng, jnp.ones(input))["params"]
    return train_state.TrainState.create(
        apply_fn=self.apply, params=params, tx=optimizer
    )



# Copyright 2022 The OTT Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import math
from typing import Callable, NamedTuple, Optional, Tuple, Union

import jax
import jax.numpy as jnp
from typing_extensions import Literal

from ott.geometry import costs, pointcloud
from ott.math import fixed_point_loop

__all__ = ["k_means", "KMeansOutput"]

Init_t = Union[Literal["k-means++", "random"],
               Callable[[pointcloud.PointCloud, int, jnp.ndarray], jnp.ndarray]]


class KPPState(NamedTuple):
  key: jnp.ndarray
  centroids: jnp.ndarray
  centroid_dists: jnp.ndarray


class KMeansState(NamedTuple):
  centroids: jnp.ndarray
  prev_assignment: jnp.ndarray
  assignment: jnp.ndarray
  errors: jnp.ndarray
  center_shift: float


class KMeansConst(NamedTuple):
  geom: pointcloud.PointCloud
  x_weights: jnp.ndarray

  @property
  def x(self) -> jnp.ndarray:
    """Array of shape ``[n, ndim]`` containing the unweighted point cloud."""
    return self.geom.x

  @property
  def weighted_x(self):
    """Array of shape ``[n, ndim]`` containing the weighted point cloud."""
    return self.x_weights[:, :-1]

  @property
  def weights(self) -> jnp.ndarray:
    """Array of shape ``[n, 1]`` containing weights for each point."""
    return self.x_weights[:, -1:]


class KMeansOutput(NamedTuple):
  """Output of the :func:`~ott.tools.k_means.k_means` algorithm.

  Args:
    centroids: Array of shape ``[k, ndim]`` containing the centroids.
    assignment: Array of shape ``[n,]`` containing the labels.
    converged: Whether the algorithm has converged.
    iteration: The number of iterations run.
    error: (Weighted) sum of squared distances from each point to its closest
      center.
    inner_errors: Array of shape ``[max_iterations,]`` containing the ``error``
      at every iteration.
  """
  centroids: jnp.ndarray
  assignment: jnp.ndarray
  converged: bool
  iteration: int
  error: float
  inner_errors: Optional[jnp.ndarray]

  @classmethod
  def _from_state(
      cls,
      state: KMeansState,
      *,
      tol: float,
      store_inner_errors: bool = False
  ) -> "KMeansOutput":
    errs = state.errors
    mask = errs == -1
    error = jnp.nanmin(jnp.where(mask, jnp.nan, errs))

    assignment_same = jnp.all(state.prev_assignment == state.assignment)
    tol_satisfied = jnp.logical_or(jnp.any(mask), (errs[-2] - errs[-1]) <= tol)
    converged = jnp.logical_or(assignment_same, tol_satisfied)

    return cls(
        centroids=state.centroids,
        assignment=state.assignment.astype(int),
        converged=converged,
        iteration=jnp.sum(~mask),
        error=error,
        inner_errors=errs if store_inner_errors else None,
    )


def _random_init(
    geom: pointcloud.PointCloud, k: int, key: jnp.ndarray
) -> jnp.ndarray:
  ixs = jnp.arange(geom.shape[0])
  ixs = jax.random.choice(key, ixs, shape=(k,), replace=False)
  return geom.subset(ixs, None).x


def _k_means_plus_plus(
    geom: pointcloud.PointCloud,
    k: int,
    key: jnp.ndarray,
    n_local_trials: Optional[int] = None,
) -> jnp.ndarray:

  def init_fn(geom: pointcloud.PointCloud, key: jnp.ndarray) -> KPPState:
    key, next_key = jax.random.split(key, 2)
    ix = jax.random.choice(key, jnp.arange(geom.shape[0]), shape=())
    centroids = jnp.full((k, geom.cost_rank), jnp.inf).at[0].set(geom.x[ix])
    dists = geom.subset(ix, None).cost_matrix[0]
    return KPPState(key=next_key, centroids=centroids, centroid_dists=dists)

  def body_fn(
      iteration: int, const: Tuple[pointcloud.PointCloud, jnp.ndarray],
      state: KPPState, compute_error: bool
  ) -> KPPState:
    del compute_error
    key, next_key = jax.random.split(state.key, 2)
    geom, ixs = const

    # no need to normalize when `replace=True`
    probs = state.centroid_dists
    ixs = jax.random.choice(
        key, ixs, shape=(n_local_trials,), p=probs, replace=True
    )
    geom = geom.subset(ixs, None)

    candidate_dists = jnp.minimum(geom.cost_matrix, state.centroid_dists)
    best_ix = jnp.argmin(candidate_dists.sum(1))

    centroids = state.centroids.at[iteration + 1].set(geom.x[best_ix])
    centroid_dists = candidate_dists[best_ix]

    return KPPState(
        key=next_key, centroids=centroids, centroid_dists=centroid_dists
    )

  if n_local_trials is None:
    n_local_trials = 2 + int(math.log(k))
  assert n_local_trials > 0, n_local_trials

  state = init_fn(geom, key)
  constants = (geom, jnp.arange(geom.shape[0]))
  state = fixed_point_loop.fixpoint_iter(
      lambda *_, **__: True,
      body_fn,
      min_iterations=k - 1,
      max_iterations=k - 1,
      inner_iterations=1,
      constants=constants,
      state=state
  )

  return state.centroids


@functools.partial(jax.vmap, in_axes=[None, 0, 0, 0], out_axes=0)
def _reallocate_centroids(
    const: KMeansConst,
    ix: jnp.ndarray,
    centroid: jnp.ndarray,
    weight: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
  is_empty = weight <= 0.
  new_centroid = (1 - is_empty) * centroid + is_empty * const.x[ix]  # (ndim,)
  centroid_to_remove = is_empty * const.weighted_x[ix]  # (ndim,)
  weight_to_remove = is_empty * const.weights[ix]  # (1,)
  return new_centroid, jnp.concatenate([centroid_to_remove, weight_to_remove])


def _update_assignment(
    const: KMeansConst,
    centroids: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
  (x, _, *args), aux_data = const.geom.tree_flatten()
  cost_matrix = type(
      const.geom
  ).tree_unflatten(aux_data, [x, centroids] + args).cost_matrix

  assignment = jnp.argmin(cost_matrix, axis=1)
  dist_to_centers = cost_matrix[jnp.arange(len(assignment)), assignment]
  return assignment, dist_to_centers


def _update_centroids(
    const: KMeansConst, k: int, assignment: jnp.ndarray,
    dist_to_centers: jnp.ndarray
) -> jnp.ndarray:
  # TODO(michalk8):
  # cannot put `k` into `const`, see https://github.com/ott-jax/ott/issues/129
  x_weights = jax.ops.segment_sum(const.x_weights, assignment, num_segments=k)
  centroids, ws = x_weights[:, :-1], x_weights[:, -1:]

  far_ixs = jnp.argsort(dist_to_centers)[-k:][::-1]
  centroids, to_remove = _reallocate_centroids(const, far_ixs, centroids, ws)
  to_remove = jax.ops.segment_sum(
      to_remove, assignment[far_ixs], num_segments=k
  )
  centroids -= to_remove[:, :-1]
  ws -= to_remove[:, -1:]

  return centroids * jnp.where(ws > 0., 1. / ws, 1.)


@functools.partial(jax.vmap, in_axes=[0] + [None] * 9)
def _k_means(
    key: jnp.ndarray,
    geom: pointcloud.PointCloud,
    k: int,
    weights: Optional[jnp.ndarray] = None,
    init: Init_t = "k-means++",
    n_local_trials: Optional[int] = None,
    tol: float = 1e-4,
    min_iterations: int = 0,
    max_iterations: int = 300,
    store_inner_errors: bool = False,
) -> KMeansOutput:

  def init_fn(init: Init_t) -> KMeansState:
    if init == "k-means++":
      init = functools.partial(
          _k_means_plus_plus, n_local_trials=n_local_trials
      )
    elif init == "random":
      init = _random_init
    if not callable(init):
      raise TypeError(
          f"Expected `init` to be 'k-means++', 'random' "
          f"or a callable, found `{init_fn!r}`."
      )

    centroids = init(geom, k, key)
    if centroids.shape != (k, geom.cost_rank):
      raise ValueError(
          f"Expected initial centroids to have shape "
          f"`{k, geom.cost_rank}`, found `{centroids.shape}`."
      )
    n = geom.shape[0]
    # TODO(michalk8): find a better solution for the below error
    # not using floats for the assignment when `fixpoint_iter_backprop` is used:

    # .../jax/_src/dtypes.py:370:
    # .0 = <set_iterator object at 0x7f4d002a1dc0>
    # >   CUB = set.intersection(*(UB[n] for n in N))
    # E   jax._src.traceback_util.UnfilteredStackTrace:
    #   KeyError: dtype([('float0', 'V')])
    # E   The stack trace below excludes JAX-internal frames.
    # E   The preceding is the original exception that occurred, unmodified.
    prev_assignment = jnp.full((n,), -2.)
    assignment = jnp.full((n,), -1.)
    errors = jnp.full((max_iterations,), -1.)

    return KMeansState(
        centroids=centroids,
        prev_assignment=prev_assignment,
        assignment=assignment,
        center_shift=jnp.inf,
        errors=errors,
    )

  def cond_fn(iteration: int, const: KMeansConst, state: KMeansState) -> bool:
    del iteration, const
    assignment_not_same = jnp.any(state.prev_assignment != state.assignment)
    tol_not_satisfied = state.center_shift > tol
    return jnp.logical_and(assignment_not_same, tol_not_satisfied)

  def body_fn(
      iteration: int, const: KMeansConst, state: KMeansState,
      compute_error: bool
  ) -> KMeansState:
    del compute_error

    assignment, dist_to_centers = _update_assignment(const, state.centroids)
    centroids = _update_centroids(const, k, assignment, dist_to_centers)
    err = jnp.sum(const.weights[:, 0] * dist_to_centers)
    center_shift = jnp.linalg.norm(state.centroids - centroids, ord="fro") ** 2

    return KMeansState(
        centroids=centroids,
        prev_assignment=state.assignment,
        assignment=assignment.astype(float),
        center_shift=center_shift,
        errors=state.errors.at[iteration].set(err)
    )

  def finalize_fn(const: KMeansConst, state: KMeansState) -> KMeansState:
    last_iter = jnp.sum(state.errors != -1) - 1

    assignment, dist_to_centers = _update_assignment(const, state.centroids)
    err = jnp.sum(const.weights[:, 0] * dist_to_centers)

    return state._replace(
        assignment=assignment.astype(float),
        errors=state.errors.at[last_iter].set(err)
    )

  force_scan = min_iterations == max_iterations
  fixpoint_fn = (  # prefer auto-diff if possible
      fixed_point_loop.fixpoint_iter if force_scan else
      fixed_point_loop.fixpoint_iter_backprop
  )
  x_weights = jnp.hstack([weights[:, None] * geom.x, weights[:, None]])
  const = KMeansConst(geom, x_weights)

  state = fixpoint_fn(
      cond_fn,
      body_fn,
      min_iterations=min_iterations,
      max_iterations=max_iterations,
      inner_iterations=1,
      constants=const,
      state=init_fn(init)
  )
  state = jax.lax.cond(
      jnp.all(state.prev_assignment == state.assignment), (lambda _, s: s),
      finalize_fn, const, state
  )

  return KMeansOutput._from_state(
      state, tol=tol, store_inner_errors=store_inner_errors
  )


def k_means(
    geom: Union[jnp.ndarray, pointcloud.PointCloud],
    k: int,
    weights: Optional[jnp.ndarray] = None,
    init: Init_t = "k-means++",
    n_init: int = 10,
    n_local_trials: Optional[int] = None,
    tol: float = 1e-4,
    min_iterations: int = 0,
    max_iterations: int = 300,
    store_inner_errors: bool = False,
    key: Optional[jnp.ndarray] = None,
) -> KMeansOutput:
  r"""K-means clustering using Lloyd's algorithm :cite:`lloyd:82`.

  Args:
    geom: Point cloud of shape ``[n, ndim]`` to cluster. If passed as an array,
      :class:`~ott.geometry.costs.SqEuclidean` cost is assumed.
    k: The number of clusters.
    weights: The weights of input points. These weights are considered when
      computing the centroids and inertia. If ``None``, use uniform weights.
    init: Initialization method. Can be one of the following:

      - **'k-means++'** - select initial centroids that are
        :math:`\mathcal{O}(\log k)`-optimal :cite:`arthur:07`.
      - **'random'** - randomly select ``k`` points from the ``geom``.
      - :func:`callable` - a function which takes the point cloud, the number of
        clusters and a random key and returns the centroids as an array of shape
        ``[k, ndim]``.

    n_init: Number of times k-means will run with different initial seeds.
    n_local_trials: Number of local trials when ``init = 'k-means++'``.
      If ``None``, :math:`2 + \lfloor log(k) \rfloor` is used.
    tol: Relative tolerance with respect to the Frobenius norm of the centroids'
      shift between two consecutive iterations.
    min_iterations: Minimum number of iterations.
    max_iterations: Maximum number of iterations.
    store_inner_errors: Whether to store the errors (inertia) at each iteration.
    key: Random key to seed the initializations.

  Returns:
    The k-means clustering.
  """
  assert geom.shape[
      0] >= k, f"Cannot cluster `{geom.shape[0]}` points into `{k}` clusters."
  if isinstance(geom, jnp.ndarray):
    geom = pointcloud.PointCloud(geom)
  if isinstance(geom.cost_fn, costs.Cosine):
    geom = geom._cosine_to_sqeucl()
  assert geom.is_squared_euclidean

  if geom.is_online:
    # to allow materializing the cost matrix
    children, aux_data = geom.tree_flatten()
    aux_data["batch_size"] = None
    geom = type(geom).tree_unflatten(aux_data, children)

  if weights is None:
    weights = jnp.ones(geom.shape[0])
  assert weights.shape == (geom.shape[0],)

  if key is None:
    key = jax.random.PRNGKey(0)
  keys = jax.random.split(key, n_init)
  out = _k_means(
      keys, geom, k, weights, init, n_local_trials, tol, min_iterations,
      max_iterations, store_inner_errors
  )
  best_ix = jnp.argmin(out.error)
  return jax.tree_util.tree_map(lambda arr: arr[best_ix], out)



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Some utility functions for transport computation.

This module is primarily made for new users who are looking for one-liners.
For instance, solving the transport between two point clouds.

>>> x = jax.random.uniform((34, 3))
>>> y = jax.random.uniform((34, 3)) + 1
>>> ot = ott.transport.solve(x, y)
>>> Tz = ot.apply(z)

Even if the `transport.solve` sole function can support many complex use cases,
we suggest more advanced users to instantiate directly their :mod:`ott.problems`
and their :mod:`ott.solvers` for better control over the parameters.
"""

from typing import Any, NamedTuple, Optional, Union

import jax.numpy as jnp
import numpy as np
from typing_extensions import Literal

from ott.geometry import geometry, pointcloud
from ott.problems.linear import linear_problem
from ott.problems.quadratic import quadratic_problem
from ott.solvers.linear import sinkhorn
from ott.solvers.quadratic import gromov_wasserstein

__all__ = ["Transport"]


class Transport(NamedTuple):
  """Transport interface to transport solutions."""

  problem: Any = None
  solver_output: Any = None

  @property
  def linear(self) -> bool:
    return isinstance(self.problem, linear_problem.LinearProblem)

  @property
  def geom(self) -> geometry.Geometry:
    return self.problem.geom if self.linear else self.solver_output.geom

  @property
  def a(self) -> jnp.ndarray:
    return self.problem.a

  @property
  def b(self) -> jnp.ndarray:
    return self.problem.b

  @property
  def linear_output(self):
    out = self.solver_output
    return out if self.linear else out.linear_state

  @property
  def matrix(self) -> jnp.ndarray:
    return self.linear_output.matrix

  def apply(self, inputs: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
    return self.linear_output.apply(inputs, axis)

  def marginal(self, axis: int = 0) -> jnp.ndarray:
    return self.linear_output.marginal(axis)


def solve(
    *args: Any,
    a: Optional[jnp.ndarray] = None,
    b: Optional[jnp.ndarray] = None,
    init_dual_a: Optional[jnp.ndarray] = None,
    init_dual_b: Optional[jnp.ndarray] = None,
    objective: Optional[Literal['linear', 'quadratic', 'fused']] = None,
    **kwargs: Any
) -> Transport:
  """Generic interface to transport problem.

  The geometries can be passed as arrays, geometry.Geometry or directly as a
  problem. The solver is passed via kwargs.

  Args:
    args: can be either a single argument, the geometry.Geometry instance, or
      for convenience only two jnp.ndarray<float> corresponding to two point
      clouds. In that case the regularization parameter epsilon must be set in
      the kwargs.
    a: the weights of the source. Uniform by default.
    b: the weights of the target. Uniform by default.
    objective: Optional[str], 'linear', 'quadratic', 'fused' or None. None
      means that the objective will be chosen based on the dimensionalities
      of the arrays.
    kwargs: the keyword arguments passed to the point clouds and/or the
      solvers.

  Returns:
    A Transport object.
  """  # noqa: D401
  tau_a, tau_b = kwargs.get('tau_a', 1.0), kwargs.get('tau_b', 1.0)
  gw_unbalanced_correction = kwargs.pop('gw_unbalanced_correction', True)
  fused_penalty = kwargs.pop('fused_penalty', None)
  eps_keys = ['epsilon', 'init', 'target', 'decay']
  pb_kwargs = {k: v for k, v in kwargs.items() if k in eps_keys}
  pb = make(
      *args,
      objective=objective,
      a=a,
      b=b,
      tau_a=tau_a,
      tau_b=tau_b,
      gw_unbalanced_correction=gw_unbalanced_correction,
      fused_penalty=fused_penalty,
      **pb_kwargs
  )
  linear = isinstance(pb, linear_problem.LinearProblem)
  solver_fn = sinkhorn.make if linear else gromov_wasserstein.make
  geom_keys = ['cost_fn', 'online']

  remove_keys = geom_keys + eps_keys if linear else geom_keys
  for key in remove_keys:
    kwargs.pop(key, None)
  solver = solver_fn(**kwargs)
  output = solver(pb, (init_dual_a, init_dual_b))
  return Transport(pb, output)


def make(
    *args: Union[jnp.ndarray, geometry.Geometry, linear_problem.LinearProblem,
                 quadratic_problem.QuadraticProblem],
    a: Optional[jnp.ndarray] = None,
    b: Optional[jnp.ndarray] = None,
    tau_a: float = 1.0,
    tau_b: float = 1.0,
    objective: Optional[str] = None,
    gw_unbalanced_correction: Optional[bool] = True,
    fused_penalty: Optional[float] = None,
    scale_cost: Optional[Union[bool, float, str]] = False,
    **kwargs: Any,
):
  """Make a problem from arrays, assuming PointCloud geometries."""
  if isinstance(args[0], (jnp.ndarray, np.ndarray)):
    x = args[0]
    y = args[1] if len(args) > 1 else args[0]
    if ((objective == 'linear') or
        (objective is None and x.shape[1] == y.shape[1])):  # noqa: E129
      geom_xy = pointcloud.PointCloud(x, y, **kwargs)
      return linear_problem.LinearProblem(
          geom_xy, a=a, b=b, tau_a=tau_a, tau_b=tau_b
      )
    elif ((objective == 'quadratic') or
          (objective is None and x.shape[1] != y.shape[1])):
      geom_xx = pointcloud.PointCloud(x, x, **kwargs)
      geom_yy = pointcloud.PointCloud(y, y, **kwargs)
      return quadratic_problem.QuadraticProblem(
          geom_xx=geom_xx,
          geom_yy=geom_yy,
          geom_xy=None,
          scale_cost=scale_cost,
          a=a,
          b=b,
          tau_a=tau_a,
          tau_b=tau_b,
          gw_unbalanced_correction=gw_unbalanced_correction
      )
    elif objective == 'fused':
      geom_xx = pointcloud.PointCloud(x, x, **kwargs)
      geom_yy = pointcloud.PointCloud(y, y, **kwargs)
      geom_xy = pointcloud.PointCloud(x, y, **kwargs)
      return quadratic_problem.QuadraticProblem(
          geom_xx=geom_xx,
          geom_yy=geom_yy,
          geom_xy=geom_xy,
          fused_penalty=fused_penalty,
          scale_cost=scale_cost,
          a=a,
          b=b,
          tau_a=tau_a,
          tau_b=tau_b,
          gw_unbalanced_correction=gw_unbalanced_correction
      )
    else:
      raise ValueError(f'Unknown transport problem `{objective}`')
  elif isinstance(args[0], geometry.Geometry):
    if len(args) == 1:
      return linear_problem.LinearProblem(
          *args, a=a, b=b, tau_a=tau_a, tau_b=tau_b
      )
    return quadratic_problem.QuadraticProblem(
        *args, a=a, b=b, tau_a=tau_a, tau_b=tau_b, scale_cost=scale_cost
    )
  elif isinstance(
      args[0],
      (linear_problem.LinearProblem, quadratic_problem.QuadraticProblem)
  ):
    return args[0]
  else:
    raise ValueError('Cannot instantiate a transport problem.')



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Soft sort operators."""

import functools
from typing import Any, Callable, Optional

import jax
import jax.numpy as jnp
import numpy as np

from ott.tools import transport


def transport_for_sort(
    inputs: jnp.ndarray,
    weights: jnp.ndarray,
    target_weights: jnp.ndarray,
    squashing_fun: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
    epsilon: float = 1e-2,
    **kwargs: Any,
) -> jnp.ndarray:
  r"""Solve reg. OT, from inputs to a weighted family of increasing values.

  Args:
    inputs: jnp.ndarray[num_points]. Must be one dimensional.
    weights: jnp.ndarray[num_points]. Weight vector `a` for input values.
    target_weights: jnp.ndarray[num_targets]: Weight vector of the target
      measure. It may be of different size than `weights`.
    squashing_fun: function taking an array to squash all its entries in [0,1].
      sigmoid of whitened values by default. Can be set to be the identity by
      passing ``squashing_fun = lambda x : x`` instead.
    epsilon: the regularization parameter.
    kwargs: keyword arguments for `sinkhorn` and / or `PointCloud`.

  Returns:
    A jnp.ndarray<float> num_points x num_target transport matrix, from all
    inputs onto the sorted target.
  """
  shape = inputs.shape
  if len(shape) > 2 or (len(shape) == 2 and shape[1] != 1):
    raise ValueError(
        f'Shape ({shape}) not supported. The input should be one-dimensional.'
    )

  x = jnp.expand_dims(jnp.squeeze(inputs), axis=1)
  if squashing_fun is None:
    squashing_fun = lambda z: jax.nn.sigmoid((z - jnp.mean(z)) /
                                             (jnp.std(z) + 1e-10))
  x = squashing_fun(x)
  a = jnp.squeeze(weights)
  b = jnp.squeeze(target_weights)
  num_targets = b.shape[0]
  y = jnp.linspace(0.0, 1.0, num_targets)[:, jnp.newaxis]

  return transport.solve(x, y, a=a, b=b, epsilon=epsilon, **kwargs)


def apply_on_axis(op, inputs, axis, *args, **kwargs: Any) -> jnp.ndarray:
  """Apply a differentiable operator on a given axis of the input.

  Args:
    op: a differentiable operator (can be ranks, quantile, etc.)
    inputs: jnp.ndarray<float> of any shape.
    axis: the axis (int) or tuple of ints on which to apply the operator. If
      several axes are passed the operator, those are merged as a single
      dimension.
    args: other positional arguments to the operator.
    kwargs: other positional arguments to the operator.

  Returns:
    A jnp.ndarray holding the output of the differentiable operator on the given
    axis.
  """
  op_inner = functools.partial(op, **kwargs)
  axis = (axis,) if isinstance(axis, int) else axis
  num_points = np.prod(np.array(inputs.shape)[(axis,)])
  permutation = np.arange(len(inputs.shape))
  axis = tuple(permutation[a] for a in axis)
  permutation = tuple(sorted(set(permutation) - set(axis)) + sorted(axis))
  inputs = jnp.transpose(inputs, permutation)

  batch_fn = jax.vmap(op_inner, in_axes=(0,) + (None,) * len(args))
  result = batch_fn(jnp.reshape(inputs, (-1, num_points)), *args)
  shrink = len(axis)
  result = jnp.reshape(result, inputs.shape[:-shrink] + result.shape[-1:])

  permutation = tuple(range(len(result.shape)))
  rank = len(result.shape) - 1
  axis = min(axis)
  permutation = permutation[:axis] + (rank,) + permutation[axis:-1]
  result = jnp.transpose(result, permutation)
  return result


def _sort(
    inputs: jnp.ndarray, topk: int, num_targets: Optional[int], **kwargs: Any
) -> jnp.ndarray:
  """Apply the soft sort operator on a one dimensional array."""
  num_points = inputs.shape[0]
  a = jnp.ones((num_points,)) / num_points
  if 0 < topk < num_points:
    start_index = 1
    b = jnp.concatenate([
        jnp.array([(num_points - topk) / num_points]),
        jnp.ones(topk, dtype=inputs.dtype) / num_points
    ])
  else:
    num_targets = num_points if num_targets is None else num_targets
    start_index = 0
    b = jnp.ones((num_targets,)) / num_targets
  ot = transport_for_sort(inputs, a, b, **kwargs)
  out = 1.0 / b * ot.apply(inputs, axis=0)
  return out[start_index:]


def sort(
    inputs: jnp.ndarray,
    axis: int = -1,
    topk: int = -1,
    num_targets: Optional[int] = None,
    **kwargs: Any,
) -> jnp.ndarray:
  r"""Apply the soft sort operator on a given axis of the input.

  Args:
    inputs: jnp.ndarray<float> of any shape.
    axis: the axis on which to apply the operator.
    topk: if set to a positive value, the returned vector will only contain
      the topk values. This also reduces the complexity of soft sorting.
    num_targets: if topk is not specified, num_targets defines the number of
      (composite) sorted values computed from the inputs (each value is a convex
      combination of values recorded in the inputs, provided in increasing
      order). If not specified, ``num_targets`` is set by default to be the size
      of the slices of the input that are sorted, i.e. the number of composite
      sorted values is equal to that of the inputs that are sorted.
    kwargs: keyword arguments passed on to lower level functions. Of interest
      to the user are ``squashing_fun``, which will redistribute the values in
      ``inputs`` to lie in [0,1] (sigmoid of whitened values by default) to
      solve the optimal transport problem; ``cost_fn``, used in ``PointCloud``,
      that defines the ground cost function to transport from ``inputs`` to the
      ``num_targets`` target values (squared Euclidean distance by default, see
      ``pointcloud.py`` for more details); ``epsilon`` values as well as other
      parameters to shape the ``sinkhorn`` algorithm.

  Returns:
    A jnp.ndarray of the same shape as the input with soft sorted values on the
    given axis.
  """
  return apply_on_axis(_sort, inputs, axis, topk, num_targets, **kwargs)


def _ranks(inputs: jnp.ndarray, num_targets, **kwargs: Any) -> jnp.ndarray:
  """Apply the soft ranks operator on a one dimensional array."""
  num_points = inputs.shape[0]
  num_targets = num_points if num_targets is None else num_targets
  a = jnp.ones((num_points,)) / num_points
  b = jnp.ones((num_targets,)) / num_targets
  ot = transport_for_sort(inputs, a, b, **kwargs)
  out = 1.0 / a * ot.apply(jnp.arange(num_targets), axis=1)
  return jnp.reshape(out, inputs.shape)


def ranks(
    inputs: jnp.ndarray,
    axis: int = -1,
    num_targets: Optional[int] = None,
    **kwargs: Any,
) -> jnp.ndarray:
  r"""Apply the soft trank operator on input tensor.

  Args:
    inputs: a jnp.ndarray<float> of any shape.
    axis: the axis on which to apply the soft ranks operator.
    num_targets: num_targets defines the number of targets used to compute a
      composite ranks for each value in ``inputs``: that soft rank will be a
      convex combination of values in [0,...,``(num_targets-2)/num_targets``,1]
      specified by the optimal transport between values in ``inputs`` towards
      those values. If not specified, ``num_targets`` is set by default to be
      the size of the slices of the input that are sorted.
    kwargs: keyword arguments passed on to lower level functions. Of interest
      to the user are ``squashing_fun``, which will redistribute the values in
      ``inputs`` to lie in [0,1] (sigmoid of whitened values by default) to
      solve the optimal transport problem; ``cost_fn``, used in ``PointCloud``,
      that defines the ground cost function to transport from ``inputs`` to the
      ``num_targets`` target values (squared Euclidean distance by default, see
      ``pointcloud.py`` for more details); ``epsilon`` values as well as other
      parameters to shape the ``sinkhorn`` algorithm.

  Returns:
    A jnp.ndarray<float> of the same shape as inputs, with the ranks.
  """
  return apply_on_axis(_ranks, inputs, axis, num_targets, **kwargs)


def quantile(
    inputs: jnp.ndarray,
    axis: int = -1,
    level: float = 0.5,
    weight: float = 0.05,
    **kwargs: Any,
) -> jnp.ndarray:
  r"""Apply the soft quantile operator on the input tensor.

  For instance:

  x = jax.random.uniform(rng, (1000,))
  q = quantile(x, 0.5, 0.01)

  Then q will be computed as a mean over the 10 median points of x.
  Therefore, there is a tradeoff between accuracy and gradient.

  Args:
   inputs: a jnp.ndarray<float> of any shape.
   axis: the axis on which to apply the operator.
   level: the value of the quantile level to be computed. 0.5 for median.
   weight: the weight of the quantile in the transport problem.
   kwargs: keyword arguments passed on to lower level functions. Of interest
      to the user are ``squashing_fun``, which will redistribute the values in
      ``inputs`` to lie in [0,1] (sigmoid of whitened values by default) to
      solve the optimal transport problem; ``cost_fn``, used in ``PointCloud``,
      that defines the ground cost function to transport from ``inputs`` to the
      ``num_targets`` target values (squared Euclidean distance by default, see
      ``pointcloud.py`` for more details); ``epsilon`` values as well as other
      parameters to shape the ``sinkhorn`` algorithm.
  Returns:
    A jnp.ndarray, which has the same shape as the input, except on the give
    axis on which the dimension is 1.
  """

  def _quantile(
      inputs: jnp.ndarray, level: float, weight: float, **kwargs
  ) -> jnp.ndarray:
    # TODO(cuturi,oliviert) option to compute several quantiles at once
    num_points = inputs.shape[0]
    a = jnp.ones((num_points,)) / num_points
    b = jnp.array([level - weight / 2, weight, 1.0 - weight / 2 - level])
    ot = transport_for_sort(inputs, a, b, **kwargs)
    out = 1.0 / b * ot.apply(jnp.squeeze(inputs), axis=0)
    return out[1:2]

  return apply_on_axis(_quantile, inputs, axis, level, weight, **kwargs)


def _quantile_normalization(
    inputs: jnp.ndarray, targets: jnp.ndarray, weights: float, **kwargs: Any
) -> jnp.ndarray:
  """Apply soft quantile normalization on a one dimensional array."""
  num_points = inputs.shape[0]
  a = jnp.ones((num_points,)) / num_points
  ot = transport_for_sort(inputs, a, weights, **kwargs)
  return 1.0 / a * ot.apply(targets, axis=1)


def quantile_normalization(
    inputs: jnp.ndarray,
    targets: jnp.ndarray,
    weights: Optional[jnp.ndarray] = None,
    axis: int = -1,
    **kwargs
) -> jnp.ndarray:
  r"""Renormalize inputs so that its quantiles match those of targets/weights.

  The idea of quantile normalization is to map the inputs to values so that the
  distribution of transformed values matches the distribution of target values.
  In a sense, we want to keep the inputs in the same order, but apply the values
  of the target.

  Args:
    inputs: the inputs array of any shape.
    targets: the target values of dimension 1. The targets must be sorted.
    weights: if set, the weights or the target.
    axis: the axis along which to apply the transformation on the inputs.
    kwargs: keyword arguments passed on to lower level functions. Of interest
      to the user are ``squashing_fun``, which will redistribute the values in
      ``inputs`` to lie in [0,1] (sigmoid of whitened values by default) to
      solve the optimal transport problem; ``cost_fn``, used in ``PointCloud``,
      that defines the ground cost function to transport from ``inputs`` to the
      ``num_targets`` target values (squared Euclidean distance by default, see
      ``pointcloud.py`` for more details); ``epsilon`` values as well as other
      parameters to shape the ``sinkhorn`` algorithm.
  Returns:
    A jnp.ndarray, which has the same shape as the input, except on the give
    axis on which the dimension is 1.

  Raises:
    A ValueError in case the weights and the targets are both set and not of
    compatible shapes.
  """
  if weights is not None and weights.shape != targets.shape:
    raise ValueError(
        'The target weights and targets values should have the '
        f'same shape: {targets.shape} != {weights.shape}'
    )
  if weights is None:
    num_targets = targets.shape[0]
    weights = jnp.ones((num_targets,)) / num_targets

  op = _quantile_normalization
  return apply_on_axis(op, inputs, axis, targets, weights, **kwargs)


def sort_with(
    inputs: jnp.ndarray,
    criterion: jnp.ndarray,
    topk: int = -1,
    **kwargs: Any,
) -> jnp.ndarray:
  r"""Sort a multidimensional array according to a real valued criterion.

  Given ``batch`` vectors of dimension `dim`, to which, for each, a real value
  ``criterion`` is associated, this function produces ``topk`` (or
  ``batch`` if ``topk`` is set to -1, as by default) composite vectors of size
  ``dim`` that will be convex combinations of all vectors, ranked from smallest
  to largest criterion. Composite vectors with the largest indices will contain
  convex combinations of those vectors with highest criterion, vectors with
  smaller indices will contain combinations of vectors with smaller criterion.

  Args:
    inputs: the inputs as a jnp.ndarray[batch, dim].
    criterion: the values according to which to sort the inputs. It has shape
      [batch, 1].
    topk: The number of outputs to keep.
    kwargs: keyword arguments passed on to lower level functions. Of interest
      to the user are ``squashing_fun``, which will redistribute the values in
      ``inputs`` to lie in [0,1] (sigmoid of whitened values by default) to
      solve the optimal transport problem; ``cost_fn``, used in ``PointCloud``,
      that defines the ground cost function to transport from ``inputs`` to the
      ``num_targets`` target values (squared Euclidean distance by default, see
      ``pointcloud.py`` for more details); ``epsilon`` values as well as other
      parameters to shape the ``sinkhorn`` algorithm.

  Returns:
    A jnp.ndarray[batch | topk, dim].
  """
  num_points = criterion.shape[0]
  weights = jnp.ones(num_points, dtype=criterion.dtype) / num_points
  if 0 < topk < num_points:
    start_index = 1
    target_weights = jnp.concatenate([
        jnp.array([(num_points - topk) / num_points]),
        jnp.ones(topk, dtype=inputs.dtype) / num_points
    ])
  else:
    start_index = 0
    target_weights = jnp.ones((num_points,)) / num_points
  ot = transport_for_sort(criterion, weights, target_weights, **kwargs)
  # Applies the topk on each of the dimensions of the inputs.
  sort_fn = jax.vmap(
      lambda x: (1.0 / target_weights * ot.apply(x, axis=0))[start_index:],
      in_axes=(1,),
      out_axes=1
  )
  return sort_fn(inputs)


def _quantize(
    inputs: jnp.ndarray, num_levels: int, **kwargs: Any
) -> jnp.ndarray:
  """Apply the soft quantization operator on a one dimensional array."""
  num_points = inputs.shape[0]
  a = jnp.ones((num_points,)) / num_points
  b = jnp.ones((num_levels,)) / num_levels
  ot = transport_for_sort(inputs, a, b, **kwargs)
  return 1.0 / a * ot.apply(1.0 / b * ot.apply(inputs), axis=1)


def quantize(
    inputs: jnp.ndarray,
    num_levels: int = 10,
    axis: int = -1,
    **kwargs: Any,
) -> jnp.ndarray:
  r"""Soft quantizes an input according using num_levels values along axis.

  The quantization operator consists in concentrating several values around
  a few predefined ``num_levels``. The soft quantization operator proposed here
  does so by carrying out a `soft` concentration that is differentiable. The
  ``inputs`` values are first soft-sorted, resulting in ``num_levels`` values.
  In a second step, the ``inputs`` values are then provided again a composite
  value that is equal (for each) to a convex combination of those soft-sorted
  values using the transportation matrix. As the regularization parameter
  ``epsilon`` of regularized optimal transport goes to 0, this operator recovers
  the expected behavior of quantization, namely each value in ``inputs`` is
  assigned a single level. When using ``epsilon>0`` the bheaviour is similar but
  differentiable.

  Args:
    inputs: the inputs as a jnp.ndarray[batch, dim].
    num_levels: number of levels available to quantize the signal.
    axis: axis along which quantization is carried out.
    kwargs: keyword arguments passed on to lower level functions. Of interest
      to the user are ``squashing_fun``, which will redistribute the values in
      ``inputs`` to lie in [0,1] (sigmoid of whitened values by default) to
      solve the optimal transport problem; ``cost_fn``, used in ``PointCloud``,
      that defines the ground cost function to transport from ``inputs`` to the
      ``num_targets`` target values (squared Euclidean distance by default, see
      ``pointcloud.py`` for more details); ``epsilon`` values as well as other
      parameters to shape the ``sinkhorn`` algorithm.

  Returns:
    A jnp.ndarray of the same size as ``inputs``.
  """
  return apply_on_axis(_quantize, inputs, axis, num_levels, **kwargs)



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A model for to embed structured features."""

from typing import Any, Tuple

import flax.linen as nn
import jax.numpy as jnp


class FeaturesEncoder(nn.Module):
  """Encodes structured features."""

  input_dims: Tuple[int]
  embed_dim: int = 32

  @nn.compact
  def __call__(self, x):
    result = []
    index = 0
    for d in self.input_dims:
      arr = x[..., index:index + d]
      result.append(arr if d == 1 else nn.Dense(self.embed_dim)(arr))
      index += d
    return jnp.concatenate(result, axis=-1)


class AdultModel(nn.Module):
  """A model to predict if the income is above 50k (adult dataset)."""

  encoder_cls: Any
  hidden: Tuple[int] = (64, 64)

  @nn.compact
  def __call__(self, x, train: bool = True):
    x = self.encoder_cls()(x)
    for h in self.hidden:
      x = nn.Dense(h)(x)
      x = nn.relu(x)
    x = nn.Dense(1)(x)
    x = nn.sigmoid(x)
    return x[..., 0]



import functools
from typing import TYPE_CHECKING, Optional, Union

import jax
import jax.experimental.sparse as jesp
import jax.numpy as jnp

if TYPE_CHECKING:
  from ott.geometry import costs

__all__ = [
    "safe_log", "kl", "js", "sparse_scale", "logsumexp",
    "barycentric_projection"
]

# TODO(michalk8): move to typing.py when refactoring types
Sparse_t = Union[jesp.CSR, jesp.CSC, jesp.COO, jesp.BCOO]


def safe_log(x: jnp.ndarray, *, eps: Optional[float] = None) -> jnp.ndarray:
  if eps is None:
    eps = jnp.finfo(x.dtype).tiny
  return jnp.where(x > 0., jnp.log(x), jnp.log(eps))


def kl(p: jnp.ndarray, q: jnp.ndarray) -> float:
  """Kullback-Leilbler divergence."""
  return jnp.vdot(p, (safe_log(p) - safe_log(q)))


def js(p: jnp.ndarray, q: jnp.ndarray, *, c: float = 0.5) -> float:
  """Jensen-Shannon divergence."""
  return c * (kl(p, q) + kl(q, p))


def sparse_scale(c: float, mat: Sparse_t) -> Sparse_t:
  """Scale a sparse matrix by a constant."""
  if isinstance(mat, jesp.BCOO):
    # most feature complete, defer to original impl.
    return c * mat
  (data, *children), aux_data = mat.tree_flatten()
  return type(mat).tree_unflatten(aux_data, [c * data] + children)


@functools.partial(jax.custom_jvp, nondiff_argnums=(1, 2, 4))
def logsumexp(mat, axis=None, keepdims=False, b=None, return_sign=False):
  return jax.scipy.special.logsumexp(
      mat, axis=axis, keepdims=keepdims, b=b, return_sign=return_sign
  )


@logsumexp.defjvp
def logsumexp_jvp(axis, keepdims, return_sign, primals, tangents):
  """Custom derivative rule for lse that does not blow up with -inf.

  This logsumexp implementation uses the standard jax one in forward mode but
  implements a custom rule to differentiate. Given the preference of jax for
  jvp over vjp, and the fact that this is a simple linear rule, jvp is used.
  This custom differentiation address issues when the output of lse is
  -inf (which corresponds to the case where all inputs in a slice are -inf,
  which happens typically when ``a`` or ``b`` weight vectors have zeros.)

  Although both exp(lse) and its derivative should be 0, automatic
  differentiation returns a NaN derivative because of a -inf - (-inf) operation
  appearing in the definition of centered_exp below. This is corrected in the
  implementation below.

  Args:
    axis: argument from original logsumexp
    keepdims: argument from original logsumexp
    return_sign: argument from original logsumexp
    primals: mat and b, the two arguments against which we differentiate.
    tangents: of same size as mat and b.

  Returns:
    original primal outputs + their tangent.
  """  # noqa: D401
  mat, b = primals
  tan_mat, tan_b = tangents
  lse = logsumexp(mat, axis, keepdims, b, return_sign)
  if return_sign:
    lse, sign = lse
  lse = jnp.where(jnp.isfinite(lse), lse, 0.0)
  centered_exp = jnp.exp(mat - jnp.expand_dims(lse, axis=axis))

  if b is None:
    res = jnp.sum(centered_exp * tan_mat, axis=axis, keepdims=keepdims)
  else:
    res = jnp.sum(b * centered_exp * tan_mat, axis=axis, keepdims=keepdims)
    res += jnp.sum(tan_b * centered_exp, axis=axis, keepdims=keepdims)
  if return_sign:
    return (lse, sign), (sign * res, jnp.zeros_like(sign))
  else:
    return lse, res


@functools.partial(jax.vmap, in_axes=[0, 0, None])
def barycentric_projection(
    matrix: jnp.ndarray, y: jnp.ndarray, cost_fn: "costs.CostFn"
) -> jnp.ndarray:
  return jax.vmap(cost_fn.barycenter, in_axes=[0, None])(matrix, y)



try:
  from importlib_metadata import PackageNotFoundError, version  # Python < 3.8
except ImportError:
  from importlib.metadata import PackageNotFoundError, version

try:
  __version__ = version("ott-jax")
except PackageNotFoundError:
  __version__ = ""

del version, PackageNotFoundError



from typing import Any, Dict, Optional, Sequence, Tuple, Union

import jax
import jax.experimental.sparse as jesp
import jax.numpy as jnp
from typing_extensions import Literal

from ott.geometry import geometry
from ott.math import decomposition, fixed_point_loop
from ott.math import utils as mu

__all__ = ["Graph"]

Sparse_t = Union[jesp.CSR, jesp.CSC, jesp.COO, jesp.BCOO]


@jax.tree_util.register_pytree_node_class
class Graph(geometry.Geometry):
  r"""Graph distance approximation using heat kernel :cite:`heitz:21,crane:13`.

  Approximates the heat kernel for large ``n_steps``, which for small ``t``
  approximates the geodesic exponential kernel :math:`e^{\frac{-d(x, y)^2}{t}}`.

  For sparse graphs, :mod:`sksparse.cholmod` is required to compute the Cholesky
  decomposition.
  Differentiating w.r.t. the edge weights is currently possible only when the
  graph is represented as a dense adjacency matrix.

  Args:
    graph: Graph represented as an adjacency matrix of shape ``[n, n]``.
      If `None`, the symmetric graph Laplacian has to be specified.
    laplacian: Symmetric graph Laplacian. The check for symmetry is **NOT**
      performed. If `None`, the graph has to be specified instead.
    t: Constant used when approximating the geodesic exponential kernel.
      If `None`, use :math:`\frac{1}{|E|} \sum_{(u, v) \in E} weight(u, v)`
      :cite:`crane:13`. In this case, the ``graph`` must be specified
      and the edge weights are all assumed to be positive.
    n_steps: Maximum number of steps used to approximate the heat kernel.
    numerical_scheme: Numerical scheme used to solve the heat diffusion.
    directed: Whether the ``graph`` is directed. If not, it will be made
      undirected as :math:`G + G^T`. This parameter is ignored when  directly
      passing the Laplacian, which is assumed to be symmetric.
    tol: Relative tolerance with respect to the Hilbert metric, see
      :cite:`peyre:19`, Remark 4.12. Used when iteratively updating scalings.
      If negative, this option is ignored and only ``n_steps`` is used.
    kwargs: Keyword arguments for :class:`~ott.geometry.geometry.Geometry`.
  """

  def __init__(
      self,
      graph: Optional[Union[jnp.ndarray, jesp.BCOO]] = None,
      laplacian: Optional[Union[jnp.ndarray, Sparse_t]] = None,
      t: Optional[float] = 1e-3,
      n_steps: int = 100,
      numerical_scheme: Literal["backward_euler",
                                "crank_nicolson"] = "backward_euler",
      directed: bool = False,
      tol: float = -1.,
      **kwargs: Any
  ):
    assert ((graph is None and laplacian is not None) or
            (laplacian is None and graph is not None)), \
           "Please provide a graph or a symmetric graph Laplacian."
    # arbitrary epsilon; can't use `None` as `mean_cost_matrix` would be used
    super().__init__(epsilon=1., **kwargs)
    self._graph = graph
    self._laplacian = laplacian
    self._solver: Optional[decomposition.CholeskySolver] = None

    self._t = t
    self.n_steps = n_steps
    self.numerical_scheme = numerical_scheme
    self.directed = directed
    self._tol = tol

  def apply_kernel(
      self,
      scaling: jnp.ndarray,
      eps: Optional[float] = None,
      axis: int = 0,
  ) -> jnp.ndarray:

    def conf_fn(
        iteration: int, solver_lap: Tuple[decomposition.CholeskySolver,
                                          Optional[Union[jnp.ndarray,
                                                         Sparse_t]]],
        old_new: Tuple[jnp.ndarray, jnp.ndarray]
    ) -> bool:
      del iteration, solver_lap

      x_old, x_new = old_new
      x_old, x_new = mu.safe_log(x_old), mu.safe_log(x_new)
      # center
      x_old, x_new = x_old - jnp.nanmax(x_old), x_new - jnp.nanmax(x_new)
      # Hilbert metric, see Remark 4.12 in `Computational Optimal Transport`
      f = x_new - x_old
      return (jnp.nanmax(f) - jnp.nanmin(f)) > self._tol

    def body_fn(
        iteration: int, solver_lap: Tuple[decomposition.CholeskySolver,
                                          Optional[Union[jnp.ndarray,
                                                         Sparse_t]]],
        old_new: Tuple[jnp.ndarray, jnp.ndarray], compute_errors: bool
    ) -> Tuple[jnp.ndarray, jnp.ndarray]:
      del iteration, compute_errors

      solver, scaled_lap = solver_lap
      _, b = old_new

      if self.numerical_scheme == "crank_nicolson":
        # below is a preferred way of specifying the update (albeit more FLOPS),
        # as CSR/CSC/COO matrices don't support adding a diagonal matrix now:
        # b' = (2 * I - M) @ b = (2 * I - (I + c * L)) @ b = (I - c * L) @ b
        b = b - scaled_lap @ b
      return b, solver.solve(b)

    # eps we cannot use since it would require a re-solve
    # axis we can ignore since the matrix is symmetric
    del eps, axis

    force_scan = self._tol < 0.
    fixpoint_fn = (
        fixed_point_loop.fixpoint_iter
        if force_scan else fixed_point_loop.fixpoint_iter_backprop
    )

    state = (jnp.full_like(scaling, jnp.nan), scaling)
    if self.numerical_scheme == "crank_nicolson":
      constants = self.solver, self._scaled_laplacian
    else:
      constants = self.solver, None

    return fixpoint_fn(
        cond_fn=(lambda *_, **__: True) if force_scan else conf_fn,
        body_fn=body_fn,
        min_iterations=self.n_steps if force_scan else 1,
        max_iterations=self.n_steps,
        inner_iterations=1,
        constants=constants,
        state=state,
    )[1]

  def apply_transport_from_scalings(
      self,
      u: jnp.ndarray,
      v: jnp.ndarray,
      vec: jnp.ndarray,
      axis: int = 0
  ) -> jnp.ndarray:

    def body_fn(carry: None, vec: jnp.ndarray) -> jnp.ndarray:
      if axis == 1:
        return carry, u * self.apply_kernel(v * vec, axis=axis)
      return carry, v * self.apply_kernel(u * vec, axis=axis)

    if not self.is_sparse:
      return super().apply_transport_from_scalings(u, v, vec, axis=axis)

    # we solve the triangular system's on host, but
    # batching rules are implemented only for `id_tap`, not for `call`
    if vec.ndim == 1:
      _, res = jax.lax.scan(body_fn, None, vec[None, :])
      return res[0, :]

    _, res = jax.lax.scan(body_fn, None, vec)
    return res

  @property
  def kernel_matrix(self) -> jnp.ndarray:
    n, _ = self.shape
    kernel = self.apply_kernel(jnp.eye(n))
    # force symmetry because of numerical imprecisions
    # happens when `numerical_scheme='backward_euler'` and small `t`
    return (kernel + kernel.T) * .5

  @property
  def cost_matrix(self) -> jnp.ndarray:
    return -self.t * mu.safe_log(self.kernel_matrix)

  @property
  def laplacian(self) -> Union[jnp.ndarray, Sparse_t]:
    """The graph Laplacian."""
    if self._laplacian is not None:
      return self._laplacian

    if self.is_sparse:
      n, _ = self.shape
      D, ixs = self.graph.sum(1).todense(), jnp.arange(n)
      D = jesp.BCOO((D, jnp.c_[ixs, ixs]), shape=(n, n))
    else:
      D = jnp.diag(self.graph.sum(1))

    # in the sparse case, we don't sum duplicates here because
    # we need to know `nnz` a priori for JIT (could be exposed in `__init__`)
    # instead, `ott.math.decomposition._jax_sparse_to_scipy` handles it on host
    return D - self.graph

  @property
  def t(self) -> float:
    """Constant used when approximating the geodesic exponential kernel."""
    if self._t is None:
      graph = self.graph
      assert graph is not None, "No graph was specified."
      if self.is_sparse:
        return jnp.mean(graph.data) ** 2
      return (jnp.sum(graph) / jnp.sum(graph > 0.)) ** 2
    return self._t

  @property
  def _scale(self) -> float:
    """Constant to scale the Laplacian with."""
    if self.numerical_scheme == "backward_euler":
      return self.t / (4. * self.n_steps)
    if self.numerical_scheme == "crank_nicolson":
      return self.t / (2. * self.n_steps)
    raise NotImplementedError(
        f"Numerical scheme `{self.numerical_scheme}` is not implemented."
    )

  @property
  def _scaled_laplacian(self) -> Union[float, jnp.ndarray, Sparse_t]:
    """Laplacian scaled by a constant, depending on the numerical scheme."""
    if self.is_sparse:
      return mu.sparse_scale(self._scale, self.laplacian)
    return self._scale * self.laplacian

  @property
  def _M(self) -> Union[jnp.ndarray, Sparse_t]:
    n, _ = self.shape
    scaled_lap = self._scaled_laplacian
    # CHOLMOD supports solving `A + beta * I`, we set `beta = 1.0`
    # when instantiating the solver
    return scaled_lap if self.is_sparse else scaled_lap + jnp.eye(n)

  @property
  def solver(self) -> decomposition.CholeskySolver:
    """Instantiate the Cholesky solver and compute the factorization."""
    if self._solver is None:
      # key/beta only used for sparse solver
      self._solver = decomposition.CholeskySolver.create(
          self._M, beta=1., key=hash(self)
      )
      # compute the factorization to avoid tracer leaks in `apply_kernel`
      # due to the scan/while loop
      _ = self._solver.L
    return self._solver

  @property
  def shape(self) -> Tuple[int, int]:
    arr = self._graph if self._graph is not None else self._laplacian
    return arr.shape

  @property
  def is_sparse(self) -> bool:
    """Whether :attr:`graph` or :attr:`laplacian` is sparse."""
    if self._laplacian is not None:
      return isinstance(self.laplacian, Sparse_t.__args__)
    if isinstance(self._graph, (jesp.CSR, jesp.CSC, jesp.COO)):
      raise NotImplementedError("Graph must be specified in `BCOO` format.")
    return isinstance(self._graph, jesp.BCOO)

  @property
  def graph(self) -> Optional[Union[jnp.ndarray, jesp.BCOO]]:
    """The underlying undirected graph as an adjacency matrix, if provided."""
    if self._graph is None:
      return None
    return (self._graph + self._graph.T) if self.directed else self._graph

  @property
  def is_symmetric(self) -> bool:
    # there are some numerical imprecisions, but it should be symmetric
    return True

  @property
  def dtype(self) -> jnp.dtype:
    return self._graph.dtype

  # TODO(michalk8): in future, use mixins for lse/kernel mode
  def transport_from_potentials(
      self, f: jnp.ndarray, g: jnp.ndarray
  ) -> jnp.ndarray:
    """Not implemented."""
    raise ValueError("Not implemented.")

  def apply_transport_from_potentials(
      self,
      f: jnp.ndarray,
      g: jnp.ndarray,
      vec: jnp.ndarray,
      axis: int = 0
  ) -> jnp.ndarray:
    """Not implemented."""
    raise ValueError("Not implemented.")

  def marginal_from_potentials(
      self,
      f: jnp.ndarray,
      g: jnp.ndarray,
      axis: int = 0,
  ) -> jnp.ndarray:
    """Not implemented."""
    raise ValueError("Not implemented.")

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:
    return [self._graph, self._laplacian, self.solver], {
        "t": self._t,
        "n_steps": self.n_steps,
        "numerical_scheme": self.numerical_scheme,
        "directed": self.directed,
        "tol": self._tol,
        **self._kwargs,
    }

  @classmethod
  def tree_unflatten(
      cls, aux_data: Dict[str, Any], children: Sequence[Any]
  ) -> "Graph":
    graph, laplacian, solver = children
    obj = cls(graph=graph, laplacian=laplacian, **aux_data)
    obj._solver = solver
    return obj
